"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
ENV Space Util Module

this file is a mypy nightmare due to lots of typing unions and complex types
do not trust mypy, trust the unit tests
"""
import copy
import logging
import typing
from collections import OrderedDict
from collections.abc import MutableSequence

import flatten_dict
import gymnasium
import numpy as np
from ray.rllib.utils.spaces import space_utils
from ray.rllib.utils.spaces.repeated import Repeated

from corl.libraries.property import CorlRepeated
from corl.libraries.state_dict import StateDict
from corl.libraries.units import Quantity, corl_get_ureg


class EnvSpaceUtil:
    """ENV Space Util"""

    _logger = logging.getLogger("EnvSpaceUtil")
    sample_type = OrderedDict | dict | tuple | np.ndarray | list | Quantity

    @staticmethod
    def deep_sanity_check_space_sample(  # noqa: PLR0912
        space: gymnasium.spaces.Space,
        sample: sample_type,
        key_stack: str = "",
    ) -> None:
        """Ensure space sample is consistent with space. This will give a traceback of the exact space that failed

        Parameters
        ----------
        space: gymnasium.spaces.Dict
            the space we expect the sample to conform to
        sample: sample_type
            sample that we are checking if it belongs to the given space
        key_stack: str
            string of the keys we have used when getting to this current spot in the observation_space, observation
            this is used for recursive calls do not set in the initial call to this function
        """
        if isinstance(space, gymnasium.spaces.Dict):
            if not isinstance(sample, StateDict | OrderedDict | dict):
                raise ValueError(
                    f"space{key_stack}={space} was a gymnasium.spaces.Dict type but\n"
                    f"sample{key_stack}={sample} \n"
                    f"is a {type(sample)} type and not a StateDict, OrderedDict, or dict\n"
                )
            for key, value in sample.items():
                EnvSpaceUtil.deep_sanity_check_space_sample(space.spaces[key], value, key_stack=f"{key_stack}[{key}]")
        elif isinstance(space, gymnasium.spaces.Tuple):
            if not isinstance(sample, tuple):
                raise ValueError(
                    f"space{key_stack}={space} is a gymnasium.spaces.Tuple type but\n"
                    f"sample{key_stack}={sample} \n"
                    f"is a {type(sample)} type and not a tuple type"
                )
            for idx, value in enumerate(sample):
                EnvSpaceUtil.deep_sanity_check_space_sample(space.spaces[idx], value, key_stack=f"{key_stack}[{idx}]")
        elif isinstance(space, gymnasium.spaces.Discrete):
            if not isinstance(sample, int | np.integer | np.ndarray):
                raise ValueError(
                    f"space{key_stack}={space} is a gymnasium.spaces.Discrete type but\n"
                    f"sample{key_stack}={sample} \n"
                    f"is a {type(sample)} type and not an int or np.integer or np.ndarray"
                )
            if not space.contains(sample):
                raise ValueError(
                    f"sample{key_stack} has value of {sample} however space{key_stack} has space definition of {space} {space.n}"
                )
        elif isinstance(space, gymnasium.spaces.Box):
            if not isinstance(sample, np.ndarray | list | np.floating):
                raise ValueError(
                    f"space{key_stack}={space} is a gymnasium.spaces.Box type but\n"
                    f"sample{key_stack}={sample} \n"
                    f"is a {type(sample)} type and not a np.ndarray, list, or np.float type"
                )
            if not space.contains(sample):
                sample_dtype = getattr(sample, "dtype", None)
                raise ValueError(
                    f"sample{key_stack} has value of {sample} however "
                    f"space{key_stack} has space definition of {space} {space.low} {space.high} "
                    f"space dtype is {space.dtype}, sample dtype is {sample_dtype}"
                )
        elif isinstance(space, gymnasium.spaces.MultiBinary):
            if not isinstance(sample, np.ndarray | list | np.integer):
                raise ValueError(
                    f"space{key_stack}={space} is a gymnasium.spaces.MultiBinary type but\n"
                    f"sample{key_stack}={sample} \n"
                    f"is a {type(sample)} type and not a np.ndarray, list, or np.integer type"
                )
            if not space.contains(sample):
                raise ValueError(
                    f"sample{key_stack} has value of {sample} however space{key_stack} has space definition of {space} {space.n}"
                )
        elif isinstance(space, gymnasium.spaces.MultiDiscrete):
            if not isinstance(sample, np.ndarray | list | np.integer):
                raise ValueError(
                    f"space{key_stack}={space} is a gymnasium.spaces.MultiDiscrete type but\n"
                    f"sample{key_stack}={sample} \n"
                    f"is a {type(sample)} type and not a np.ndarray, list, or np.integer type"
                )
            if not space.contains(sample):
                raise ValueError(
                    f"sample{key_stack} has value of {sample} however space{key_stack} has space definition of {space} {space.nvec}"
                )
        elif isinstance(space, Repeated):
            if not isinstance(sample, list):
                raise ValueError(
                    f"space{key_stack}={space} is a ray.rllib.utils.spaces.repeated.Repeated type but\n"
                    f"sample{key_stack}={sample}\n"
                    f"is a {type(sample)} type and not a list  type"
                )
            for idx, item in enumerate(sample):
                EnvSpaceUtil.deep_sanity_check_space_sample(space.child_space, item, key_stack=f"{key_stack}[{idx}]")

    @staticmethod
    def sanity_check_space_sample(space, sample):
        """[summary]

        Parameters
        ----------
        space : [type]
            [description]
        sample : [type]
            [description]

        Raises
        ------
        RuntimeError
            [description]
        """
        if not space.contains(sample):
            raise RuntimeError(f"sample of {sample} does not meet space {space} setup")

    @staticmethod
    def deep_merge_dict(source: dict, destination: dict):
        """
        Merge two dictionaries that also can contain sub dictionaries. This function returns the second dict
        but it also modifies it in place

        run me with nosetests --with-doctest file.py

        >>> a = { 'first' : { 'all_rows' : { 'pass' : 'dog', 'number' : '1' } } }
        >>> b = { 'first' : { 'all_rows' : { 'fail' : 'cat', 'number' : '5' } } }
        >>> merge(b, a) == { 'first' : { 'all_rows' : { 'pass' : 'dog', 'fail' : 'cat', 'number' : '5' } } }
        True
        """
        for key, value in source.items():
            if isinstance(value, dict):
                # get node or create one
                node = destination.setdefault(key, OrderedDict())
                EnvSpaceUtil.deep_merge_dict(value, node)
            else:
                destination[key] = value

        return destination

    @staticmethod
    def scale_space(space: gymnasium.spaces.Space, scale: float) -> gymnasium.spaces.Space:
        """
        Multiplies the low and high properties of all the Boxes in the given gymnasium space by the scale input
        Parameters
        ----------
        space: gymnasium.spaces.Space
            the gymnasium space to scale the Boxes of
        scale: float
            what to multiply the Box low and high by
        Returns
        -------
        gymnasium.spaces.Space
            the scaled gymnasium space
        """
        # TODO: this copy probably doesn't actually work but I can dream
        val = copy.deepcopy(space)
        if isinstance(space, gymnasium.spaces.Dict):
            new_dict = OrderedDict()
            for key, value in space.spaces.items():
                new_dict[key] = EnvSpaceUtil.scale_space(value, scale=scale)
            val = gymnasium.spaces.Dict(spaces=new_dict)
        elif isinstance(space, gymnasium.spaces.Tuple):
            new_thing = [EnvSpaceUtil.scale_space(sp, scale=scale) for sp in space.spaces]
            val = gymnasium.spaces.Tuple(tuple(new_thing))
        elif isinstance(space, gymnasium.spaces.Box):
            scaled_box = gymnasium.spaces.Box(
                low=np.multiply(space.low, scale).astype(np.float32),
                high=np.multiply(space.high, scale).astype(np.float32),
                shape=space.shape,
                dtype=np.float32,
            )
            val = scaled_box

        return val

    @staticmethod
    def zero_mean_space(space: gymnasium.spaces.Space) -> gymnasium.spaces.Space:
        """
        Returns a space object where every Box instance has its low and high shifted to be zero mean
        Parameters
        ----------
        space: gymnasium.spaces.Space
            The gymnasium space to zero mean

        Returns
        -------
        gymnasium.spaces.Space
            A gymnasium space the same as the input but with the Box instances shifted
        """
        # TODO: this copy doesn't actually work but I can dream
        val = copy.deepcopy(space)
        if isinstance(space, gymnasium.spaces.Dict):
            new_dict = OrderedDict()
            for key, value in space.spaces.items():
                new_dict[key] = EnvSpaceUtil.zero_mean_space(value)
            val = gymnasium.spaces.Dict(spaces=new_dict)
        elif isinstance(space, gymnasium.spaces.Tuple):
            new_thing = [EnvSpaceUtil.zero_mean_space(sp) for sp in space.spaces]
            val = gymnasium.spaces.Tuple(tuple(new_thing))
        elif isinstance(space, gymnasium.spaces.Box):
            mean = (space.high + space.low) / 2
            zero_mean_box = gymnasium.spaces.Box(low=space.low - mean, high=space.high - mean, shape=space.shape, dtype=np.float32)
            val = zero_mean_box

        return val

    @staticmethod
    def space_box_min_maxer(
        space_likes: tuple[gymnasium.spaces.Space],
        out_min: float = -1.0,
        out_max: float = 1.0,
    ) -> gymnasium.spaces.Space:
        """
        Makes a gymnasium box to the out_min and out_max range

        Parameters
        ----------
        space_likes: typing.Tuple[gymnasium.spaces.Space]
            the gymnasium space to turn all boxes into the scaled space
        out_min: float
            the new low for the boxes
        out_max: float
            the new high for the boxes

        Returns
        -------
        gymnasium.spaces.Space:
            the new gymnasium spaces where all boxes have had their bounds changed
        """
        space_arg = space_likes[0]
        if isinstance(space_arg, gymnasium.spaces.Box) and space_arg.is_bounded():
            return gymnasium.spaces.Box(low=out_min, high=out_max, shape=space_arg.shape, dtype=np.float32)
        return copy.deepcopy(space_arg)

    @staticmethod
    def normalize_space(
        space: gymnasium.spaces.Space,
        out_min=-1,
        out_max=1,
    ) -> gymnasium.spaces.Space:
        """
        This is a convenience wrapper for box_scaler

        Parameters
        ----------
        space: gymnasium.spaces.Space
            the gymnasium space to turn all boxes into the scaled space
        out_min: float
            the new low for the boxes
        out_max: float
            the new high for the boxes

        Returns
        -------
        gymnasium.spaces.Space:
            the new gymnasium spaces where all boxes have had their bounds changed
        """
        return EnvSpaceUtil.iterate_over_space(
            func=EnvSpaceUtil.space_box_min_maxer,
            space=space,
            out_min=out_min,
            out_max=out_max,
        )

    @staticmethod
    def get_zero_sample_from_space(space: gymnasium.spaces.Space) -> sample_type:
        """
        Given a gymnasium space returns an instance of that space but instead of sampling from the
        gymnasium space, returns all zeros. If the space is not a Box and we cannot iterate over it
        then we will sample from it.

        Parameters
        ----------
        space: gymnasium.spaces.Space
            The gymnasium space to zero sample from

        Returns
        -------
        sample_type
            The instance of the gymnasium space but all Box spaces are sampled as zero
        """
        val = space.sample()
        if isinstance(space, gymnasium.spaces.Dict):
            new_dict = OrderedDict()
            for key, value in space.spaces.items():
                new_dict[key] = EnvSpaceUtil.get_zero_sample_from_space(value)
            val = new_dict
        elif isinstance(space, gymnasium.spaces.Tuple):
            new_tuple = [EnvSpaceUtil.get_zero_sample_from_space(sp) for sp in space.spaces]
            val = tuple(new_tuple)
        elif isinstance(space, gymnasium.spaces.Box):
            val = np.zeros(shape=space.shape, dtype=np.float32)
        elif isinstance(space, gymnasium.spaces.Discrete):
            val = 0
        return val

    @staticmethod
    def get_mean_sample_from_space(space: gymnasium.spaces.Space) -> sample_type:
        """
        Given a gymnasium space returns an instance of that space but instead of sampling from the
        gymnasium space, returns all zeros. If the space is not a Box and we cannot iterate over it
        then we will sample from it.

        Parameters
        ----------
        space: gymnasium.spaces.Space
            The gymnasium space to zero sample from

        Returns
        -------
        sample_type
            The instance of the gymnasium space but all Box spaces are sampled as zero
        """
        val = space.sample()
        if isinstance(space, gymnasium.spaces.Dict):
            new_dict = OrderedDict()
            for key, value in space.spaces.items():
                new_dict[key] = EnvSpaceUtil.get_mean_sample_from_space(value)
            val = new_dict
        elif isinstance(space, gymnasium.spaces.Tuple):
            new_tuple = [EnvSpaceUtil.get_mean_sample_from_space(sp) for sp in space.spaces]
            val = tuple(new_tuple)
        elif isinstance(space, gymnasium.spaces.Box):
            val = (space.high + space.low) / 2.0
        return val

    @staticmethod
    def add_space_samples(
        space_template: gymnasium.spaces.Space,
        space_sample1: sample_type,
        space_sample2: sample_type,
    ) -> sample_type:
        """
        Adds together two instances of gymnasium spaces. This only adds the ndarray or list types (that were sampled from Box)
        If the object is not a ndarray or list then the value of space_sample1 is returned by default
        Parameters
        ----------
        space_template: gymnasium.spaces.Space
            The template to use for adding these space instances.
            This is to determine the difference between a Box and Discrete or MultiDiscrete or MultiBinary
        space_sample1: sample_type
            The first instance to add
        space_sample2: sample_type
            The second instance to add
        Returns
        -------
        sample_type
            an instance of the space object but with all the Box types added

        """
        # if not type(space_sample1) == type(space_sample2):
        #     raise ValueError('space instances must be of same type')
        # TODO: I want to check they are the same type but dict and OrderedDict should match which makes this annoying
        val: EnvSpaceUtil.sample_type
        if isinstance(space_template, gymnasium.spaces.Dict):
            assert isinstance(space_sample1, dict)
            if isinstance(space_sample2, dict):
                new_dict = OrderedDict()
                for key, space_value in space_template.spaces.items():
                    value1 = space_sample1[key]
                    value2 = space_sample2[key]
                    new_dict[key] = EnvSpaceUtil.add_space_samples(space_value, value1, value2)
                val = new_dict
            elif isinstance(space_sample2, Quantity):
                new_dict = OrderedDict()
                for key, space_value in space_template.spaces.items():
                    value1 = space_sample1[key]
                    new_dict[key] = EnvSpaceUtil.add_space_samples(space_value, value1, space_sample2)
                val = new_dict
            else:
                raise ValueError("space_sample2 needs to be a dict or a quantity")
        elif isinstance(space_template, gymnasium.spaces.Tuple):
            assert isinstance(space_sample1, tuple)
            assert isinstance(space_sample2, tuple)
            new_tuple = [EnvSpaceUtil.add_space_samples(*args) for args in zip(space_template, space_sample1, space_sample2)]
            val = tuple(new_tuple)
        elif isinstance(space_template, gymnasium.spaces.Box):
            if isinstance(space_sample1, np.ndarray):
                val = np.array(space_sample1 + space_sample2)
            elif isinstance(space_sample1, list):
                assert isinstance(space_sample2, list)
                val = [value1 + value2 for value1, value2 in zip(space_sample1, space_sample2)]
            elif isinstance(space_sample1, Quantity):
                assert isinstance(space_sample2, Quantity)
                val = space_sample1.m + space_sample2.m
                val = corl_get_ureg().Quantity(val, space_sample1.u)
        else:
            val = copy.deepcopy(space_sample1)
        return val

    @staticmethod
    def clip_space_sample_to_space(space_sample: sample_type, space: gymnasium.Space, is_wrap: bool = False) -> sample_type:
        """
        Clips a space instance to a given space. After this the space should contain the space instance

        Parameters
        ----------
        space_sample: sample_type
            the space instance we are going to clip
        space: gymnasium.spaces.Space
            the gymnasium space to clip the instance to.
        Returns
        -------
        sample_type
            the clipped space instance
        """
        val = space_sample
        if isinstance(space, Repeated):
            assert isinstance(val, list)
            new_obs = [EnvSpaceUtil.clip_space_sample_to_space(item, space.child_space, is_wrap=is_wrap) for item in val]

            val = new_obs
        if isinstance(space, gymnasium.spaces.Dict):
            assert isinstance(space_sample, dict)
            new_dict = OrderedDict()
            for key, space_value in space.spaces.items():
                space_sample_value = space_sample[key]
                new_dict[key] = EnvSpaceUtil.clip_space_sample_to_space(space_sample_value, space_value, is_wrap)
            val = new_dict
        elif isinstance(space, gymnasium.spaces.Tuple):
            assert isinstance(space_sample, tuple)
            new_tuple = [EnvSpaceUtil.clip_space_sample_to_space(siv, sv, is_wrap) for siv, sv in zip(space_sample, space)]
            val = tuple(new_tuple)
        elif isinstance(space, gymnasium.spaces.Box):
            assert isinstance(space_sample, Quantity)
            assert isinstance(val, Quantity)
            # A glue should always return
            val_m = val.m
            if is_wrap:
                # Takes care of the case where the controls wrap at the min/max value
                # Example: Range = 0-1, value = 1.1 ----> clipping puts to .1
                if val_m > space.high:
                    val_m = space.low + (val_m - space.high)
                elif val_m < space.low:
                    val_m = space.high - (space.low - val_m)
            else:
                # Takes care of the case where the controls saturate at the min/max value
                # Example: Range = 0-1, value = 1.1 ----> clipping puts to 1
                val_m = np.clip(a=val_m, a_min=space.low, a_max=space.high)
            val = corl_get_ureg().Quantity(val_m, val.u)
        return val

    @staticmethod
    def iterate_over_space(func, space: gymnasium.spaces.Space, *func_args, **func_kwargs) -> gymnasium.spaces.Space:
        """
        utility tool for iterate over space to help mypy handle the return type correctly
        and be a little easier to use

        Args:
            func (_type_): function to call on space and all sub space, will receive
                            func_args and func_kwargs as arguments
            space (gymnasium.spaces.Space): space to call func on

        Raises:
            RuntimeError: if this function tries not to return a gymnasium.spaces.Space

        Returns:
            gymnasium.spaces.Space: Space after func has operated on it
        """
        tmp = EnvSpaceUtil.iterate_over_space_likes(func, (space,), True, *func_args, **func_kwargs)
        if isinstance(tmp, gymnasium.spaces.Space):
            return tmp
        raise RuntimeError("bug spat, should never happen, iterate_over_space tried to return something other than a gymnasium Space")

    @staticmethod
    def iterate_over_space_with_sample(func, space: gymnasium.spaces.Space, sample: sample_type, *func_args, **func_kwargs) -> sample_type:
        """
        utility tool for iterate over space and sample to help mypy handle the return type correctly
        and be a little easier to use

        Args:
            func (_type_): function to call on sample and all sub samples, will receive
                            func_args and func_kwargs as arguments
            space (gymnasium.spaces.Space): space to call func on
            sample: (sample_type): sample to call func on

        Raises:
            RuntimeError: if this function tries to not return a sample

        Returns:
            sample: space sample after func has operated on it
        """
        tmp = EnvSpaceUtil.iterate_over_space_likes(func, (space, sample), False, *func_args, **func_kwargs)
        if not isinstance(tmp, gymnasium.spaces.Space):
            return tmp
        raise RuntimeError("bug spat, should never happen, iterate_over_space tried to return something other than a sample_type")

    # TODO: maybe use this function in more places? maybe not? it could be slower?
    @staticmethod
    def iterate_over_space_likes(
        func,
        space_likes: tuple[gymnasium.spaces.Space | sample_type, ...],
        return_space: bool,
        *func_args,
        **func_kwargs,
    ) -> gymnasium.spaces.Space | sample_type:
        """
        Iterates over space_likes which are tuple, dicts or the gymnasium equivalent.
        When it encounters an actual item that is not a container it calls the func method.
        put any args, or kwargs you want to give to func in the overall call and they will be forwarded

        Parameters
        ----------
        func:
            the function to apply
        space_likes: typing.Tuple[typing.Union[gymnasium.spaces.Space, sample_type], ...]
            the spaces to iterate over. They must have the same keywords for dicts and number of items for tuples
        return_space: bool
            if true the containers will be gymnasium space equivalents
        func_args
            the arguments to give to func
        func_kwargs
            the keyword arguments to give to func

        Returns
        -------
        The contained result by calling func and stuffing back into the tuples and dicts in the call
        if return_space=True the containers are gymnasium spaces
        """

        first_space_like = space_likes[0]
        val = None
        if isinstance(first_space_like, gymnasium.spaces.Dict | dict | OrderedDict):
            new_dict = OrderedDict()
            keys: typing.KeysView
            keys = first_space_like.spaces.keys() if isinstance(first_space_like, gymnasium.spaces.Dict) else first_space_like.keys()

            for key in keys:
                # type ignoring this because mypy is really struggling with the gymnasium.Dict vs dict differences
                new_space_likes = tuple(spacer[key] for spacer in space_likes)  # type: ignore[index]
                new_dict[key] = EnvSpaceUtil.iterate_over_space_likes(  # type: ignore[misc]
                    func,
                    space_likes=new_space_likes,
                    return_space=return_space,
                    *func_args,  # noqa: B026
                    **func_kwargs,
                )
            val = gymnasium.spaces.Dict(spaces=new_dict) if return_space else new_dict  # type: ignore
        elif isinstance(first_space_like, gymnasium.spaces.Tuple | tuple):
            new_tuple = [
                EnvSpaceUtil.iterate_over_space_likes(  # type: ignore[misc]
                    func,
                    space_likes=new_space_likes,
                    return_space=return_space,
                    *func_args,  # noqa: B026
                    **func_kwargs,
                )
                for new_space_likes in zip(*space_likes)  # type: ignore
            ]
            val = gymnasium.spaces.Tuple(tuple(new_tuple)) if return_space else tuple(new_tuple)  # type: ignore
        elif isinstance(first_space_like, Repeated):
            # if space_likes is longer than 1 that means that return_space = False
            # if there is only the space that means we need to generate just the space
            # itself and can use the second path, however for the case where we have a sample
            # it comes in as a list and we must iterate over the entire repeated list and process it
            if len(space_likes) > 1:
                repeated_samples = space_likes[1]
                assert isinstance(
                    repeated_samples, MutableSequence
                ), f"repeated_samples must be MutableSequence, received {type(repeated_samples).__name__}"
                val = [  # type: ignore
                    EnvSpaceUtil.iterate_over_space_likes(  # type: ignore[misc]
                        func,
                        space_likes=(first_space_like.child_space, sample),
                        return_space=return_space,
                        *func_args,  # noqa: B026
                        **func_kwargs,
                    )
                    for sample in repeated_samples
                ]
            else:
                new_child_space = EnvSpaceUtil.iterate_over_space_likes(  # type: ignore[misc]
                    func,
                    space_likes=(first_space_like.child_space,),
                    return_space=return_space,
                    *func_args,  # noqa: B026
                    **func_kwargs,
                )
                val = CorlRepeated(child_space=new_child_space, max_len=first_space_like.max_len)  # type: ignore
        else:
            val = func(space_likes, *func_args, **func_kwargs)
        return val  # type: ignore

    @staticmethod
    def box_scaler(
        space_likes: tuple[gymnasium.spaces.Space, sample_type],
        out_min: float = -1,
        out_max: float = 1,
    ) -> sample_type:
        """
        This scales a box space to be between the out_min and out_max arguments

        Parameters
        ----------
        space_likes: typing.Tuple[gymnasium.spaces.Space, sample_type]
            the first is the gymnasium space to determine the input min and max
            the second is the sample of this space to scale
        out_min: float
            the minimum of the output scaling
        out_max: float
            the maximum of the output scaling

        Returns
        -------
        sample_type:
            the scaled sample with min of out_min and max of out_max
        """
        (space_arg, space_sample_arg) = space_likes
        if isinstance(space_arg, gymnasium.spaces.Box) and space_arg.is_bounded():
            in_min = space_arg.low
            in_max = space_arg.high
            norm_value = (out_max - out_min) * (space_sample_arg - in_min) / (in_max - in_min) + out_min
            return norm_value.astype(np.float32)
        return copy.deepcopy(space_sample_arg)

    @staticmethod
    def scale_sample_from_space(
        space: gymnasium.spaces.Space,
        space_sample: sample_type,
        out_min: float = -1,
        out_max: float = 1,
    ) -> sample_type:
        """
        This is a convenience wrapper for box_scaler

        Parameters
        ----------
        space: gymnasium.spaces.Space
            the space to use for the input min and max
        space_sample: sample_type
            the space sample to scale
        out_min: float
            the minimum of the output scaling
        out_max: float
            the maximum of the output scaling

        Returns
        -------
        sample_type:
            the scaled sample with min of out_min and max of
            out_max (this is in dicts and tuples the same as space_sample was)
        """
        return EnvSpaceUtil.iterate_over_space_with_sample(
            func=EnvSpaceUtil.box_scaler,
            space=space,
            sample=space_sample,
            out_min=out_min,
            out_max=out_max,
        )

    @staticmethod
    def box_unscaler(
        space_likes: tuple[gymnasium.spaces.Space, sample_type],
        out_min: float = -1,
        out_max: float = 1,
    ) -> sample_type:
        """
        Unscales the space_sample according to be the scale of the input space.
        In this sense out_min and out_max are the min max of the sample

        Parameters
        ----------
        space_likes: typing.Tuple[gymnasium.spaces.Space, sample_type]
            the first is the gymnasium spade to determine the input min and max
            the second is the sample of this space to scale
        out_min: float
            the minimum of the sample
        out_max: float
            the maximum of the sample

        Returns
        -------
        space_type:
            the unscaled sample
        """
        (space_arg, space_sample_arg) = space_likes
        if isinstance(space_arg, gymnasium.spaces.Box):
            norm_value = space_sample_arg
            assert isinstance(norm_value, np.ndarray), f"norm_value must be np.ndarray, received {type(norm_value).__name__}"
            in_min = space_arg.low
            in_max = space_arg.high
            return (norm_value - out_min) * (in_max - in_min) / (out_max - out_min) + in_min

        return copy.deepcopy(space_sample_arg)

    @staticmethod
    def unscale_sample_from_space(
        space: gymnasium.spaces.Space,
        space_sample: sample_type,
        out_min: float = -1,
        out_max: float = 1,
    ) -> sample_type:
        """
        This is a convenience wrapper for box_unscaler

        Parameters
        ----------
        space: gymnasium.spaces.Space
            the gymnasium space we will unscale to
        space_sample: sample_type
            the sample we want to unscale. thus this is a scaled version of the input space
            with a min,max defined by the arguments out_min,out_max
        out_min: float
            the minimum of the sample
        out_max: float
            the maximum of the sample

        Returns
        -------
        sample_type:
            the unscaled version of the space_sample. Thus the space should now contain this output sample
        """

        return EnvSpaceUtil.iterate_over_space_with_sample(
            func=EnvSpaceUtil.box_unscaler,
            space=space,
            sample=space_sample,
            out_min=out_min,
            out_max=out_max,
        )

    @staticmethod
    def convert_config_param_to_space(action_space: gymnasium.spaces.Space, parameter: int | (float | (list | dict))) -> dict:
        """
        This is a convert for parameters used in action space conversions

        Parameters
        ----------
        space: gymnasium.spaces.Space
            the gymnasium space that defines the actions
        parameter: typing.Union[int, float, list, dict]
            the parameter as defined in a a config

        Returns
        -------
        dict:
            actions are dicts, the output is a dict with the parameters set for each key
        """
        action_params = {}
        sample_action = action_space.sample().items()
        if isinstance(parameter, int | float):
            parameter = [parameter]  # change to list if needed
            if len(parameter) == 1:  # if length 1, broadcast to the proper length
                for key, value in sample_action:
                    action_params[key] = parameter * len(value)
        elif isinstance(parameter, list):
            if len(sample_action) != 1:
                raise ValueError(f"list configs can only be applied to action space dicts with single key: {sample_action}")
            for key, value in sample_action:
                if len(value) != len(parameter):
                    raise ValueError(f"config length does not match action length of {len(value)}. {parameter}")
                action_params[key] = parameter
        elif isinstance(parameter, dict):
            for key, value in sample_action:
                if key not in parameter:
                    if isinstance(key, tuple):  # parameters may be specified using tuple values as keys
                        action_params[key] = np.zeros(len(key))  # type: ignore[assignment]
                        for idx, sub_key in enumerate(key):
                            if sub_key not in parameter:
                                raise ValueError(f"action space key not in config key: {sub_key}, config: {parameter} ")
                            action_params[key][idx] = parameter[sub_key]
                    else:
                        raise ValueError(f"action space key not in config key: {key}, config: {parameter} ")
                elif len(value) != len(parameter[key]):
                    raise ValueError(f"config value length for key {key} does not match action length of {len(value)}. {parameter}")
                else:
                    action_params[key] = parameter[key]

        return action_params


def convert_gymnasium_space(input_space: gymnasium.Space, output_type: type[gymnasium.Space]) -> gymnasium.Space:
    """Converts a gymnasium space to the specified output type.

    Parameters
    ----------
    input_space : gymnasium.spaces
        The input space to convert.
    output_type : Type[gymnasium.spaces.Space]
        The desired output space type.

    Returns
    -------
    gymnasium.spaces.Space
        The converted gymnasium space.

    Raises
    ------
    ValueError
        For Discrete output types, the input must contain only one discrete gymnasium space.
    NotImplementedError
        Unsupported conversions.
    """

    if isinstance(input_space, gymnasium.spaces.Dict):
        if output_type is SingleLayerDict:
            return SingleLayerDict(input_space)

        if output_type is gymnasium.spaces.discrete.Discrete:
            flattened_space_list = space_utils.flatten_space(input_space)
            if len(flattened_space_list) > 1:
                raise ValueError(f"Too many inputs. Cannot convert: {input_space} into a single Discrete action")
            if isinstance(flattened_space_list[0], gymnasium.spaces.discrete.Discrete):
                return flattened_space_list[0]
            raise ValueError(f"Conversion invalid: {flattened_space_list[0]} -> {output_type}")

    if (
        isinstance(
            input_space, gymnasium.spaces.Dict | gymnasium.spaces.MultiDiscrete | gymnasium.spaces.Tuple | gymnasium.spaces.MultiBinary
        )
        and output_type is gymnasium.spaces.box.Box
    ):
        flattened_space = gymnasium.spaces.flatten_space(input_space)
        # Check to make sure the resultant space is a box. Raise error if not.
        if not isinstance(flattened_space, gymnasium.spaces.Box):
            raise NotImplementedError(f"{input_space} contains a space where the conversion is not implemented.")
        return flattened_space

    raise NotImplementedError(f"Conversion not implemented: {type(input_space)} -> {output_type}")


def convert_sample(sample: typing.Any, sample_space: gymnasium.spaces.Space, output_space: gymnasium.spaces.Space):
    """Converts the sample into the output_space format.

    Parameters
    ----------
    sample : typing.Any
        A sample from `sample_space`.
    sample_space : gymnasium.spaces.Space
        The space associated with the sample.
    output_space : gymnasium.spaces.Space
        The output space to convert the sample into.

    Returns
    -------
    typing.Any
        The converted sample.

    Raises
    ------
    NotImplementedError
        When an unsupported conversion is passed in.
    """
    # sample_space: Discrete -> output_space:Dict
    if isinstance(sample_space, gymnasium.spaces.Discrete) and isinstance(output_space, gymnasium.spaces.Dict):
        # Turn the discrete input value into a one-hot encoded vector
        sample = gymnasium.spaces.flatten(sample_space, sample)
        return gymnasium.spaces.unflatten(output_space, sample)

    # sample_space: Box -> output_space: Dict
    if isinstance(sample_space, gymnasium.spaces.Box) and isinstance(
        output_space, gymnasium.spaces.Dict | gymnasium.spaces.MultiDiscrete | gymnasium.spaces.Tuple | gymnasium.spaces.MultiBinary
    ):
        return gymnasium.spaces.unflatten(output_space, sample)

    if isinstance(
        sample_space, gymnasium.spaces.Dict | gymnasium.spaces.MultiDiscrete | gymnasium.spaces.Tuple | gymnasium.spaces.MultiBinary
    ) and isinstance(output_space, gymnasium.spaces.Box):
        return gymnasium.spaces.flatten(sample_space, sample)

    # sample_space: Dict -> output_space: SingleLayerDict
    if isinstance(sample_space, gymnasium.spaces.Dict) and isinstance(output_space, SingleLayerDict):
        return SingleLayerDict.flatten_and_convert_to_ordered_dict(sample)

    # sample_space: SingleLayerDict -> output_space: Dict
    if isinstance(sample_space, SingleLayerDict) and isinstance(output_space, gymnasium.spaces.Dict):
        return OrderedDict(flatten_dict.unflatten(sample, splitter="tuple"))

    if isinstance(sample_space, gymnasium.spaces.Dict) and isinstance(output_space, gymnasium.spaces.Discrete):
        flattened_sample_list = flatten_sample(sample)

        if len(flattened_sample_list) > 1:
            raise ValueError(f"Too many inputs. Cannot convert: {sample_space} into a single Discrete action")
        return flattened_sample_list[0]

    raise NotImplementedError(f"Conversion of sample is not implemented for: {sample_space}->{output_space}.")


class SingleLayerDict(gymnasium.spaces.Dict):
    """A single layered gymnasium.spaces.Dict

    Inherits gymnasium.spaces.Dict

    Parameters
    ----------
    input_space_dict : gymnasium.spaces.Dict
        Input gymnasium space dictionary to flatten.
    """

    def __init__(self, input_space_dict: gymnasium.spaces.Dict) -> None:
        converted_dict = SingleLayerDict._gymnasium_dict_to_python_dict(copy.deepcopy(input_space_dict))
        converted_dict = SingleLayerDict.flatten_and_convert_to_ordered_dict(converted_dict.spaces)
        super().__init__(converted_dict)

    @staticmethod
    def _gymnasium_dict_to_python_dict(input_space_dict: gymnasium.spaces.Dict) -> typing.Any:
        """Recursively converts all nested gymnasium.spaces.Dicts into python dictionaries

        Parameters
        ----------
        input_space_dict : gymnasium.spaces.Space
            The input gymnasium space dictionary to be converted into a python dictinoary.

        Returns
        -------
        Union[dict, gymnasium.spaces.Dict]
            Converted output.

        """
        for key, value in input_space_dict.items():
            if isinstance(value, gymnasium.spaces.Dict):
                # Converts the gymnasium.spaces.Dict -> dictionary
                gymnasium_dict = gymnasium.spaces.Dict(value.spaces)
                SingleLayerDict._gymnasium_dict_to_python_dict(gymnasium_dict)
                input_space_dict[key] = gymnasium_dict
        return input_space_dict

    @staticmethod
    def flatten_and_convert_to_ordered_dict(input_dict: dict | OrderedDict) -> OrderedDict:
        """Flattens and sorts a python dictionary.

        Parameters
        ----------
        input_dict : typing.Union[dict, OrderedDict]
            A dictionary or OrderedDict.

        Returns
        -------
        OrderedDict
            An ordered dictionary sorted by the key.
        """
        return OrderedDict(list(flatten_dict.flatten(input_dict, reducer="tuple").items()))


def flatten_sample(sample):
    """Flattens a sampe into its primitive components using rllib flatten technique.

    Primitive components are any non Tuple/Dict types.

    Args:
        sample: action or observation to flatten. This may be any
            supported type (including nested Tuples and Dicts).

    Returns:
        List[prim_data]: The flattened list of primitive types. This list
            does not contain Tuples or Dicts anymore.
    """

    def _helper_flatten(sample_, return_list):
        if isinstance(sample_, tuple):
            for s in sample_:
                _helper_flatten(s, return_list)
        elif isinstance(sample_, dict | OrderedDict):
            for k in sample_:
                _helper_flatten(sample_[k], return_list)
        else:
            return_list.append(sample_)

    ret = []
    _helper_flatten(sample, ret)
    return ret



"""
---------------------------------------------------------------------------


Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Functions related to dones for the streamlit app
"""

import pandas as pd
import plotly.express as px
import streamlit as st
from streamlit.runtime.state.session_state_proxy import SessionStateProxy

from corl.visualization.streamlit_app.utils import convert_counts_to_percentage


def platform_dones_counts_and_table(platform_dones_table: pd.DataFrame, use_container_width: bool) -> None:
    """Generates a streamlit table of platform dones (data in `platform_dones_table`)
    and barcharts grouped by platform name and dones

    Parameters
    ----------
    platform_dones_table : pd.DataFrame
        The platform dones table from EpisodeArtifactTables
    use_container_width : bool
        When True, uses the full container width (streamlit parameter)
    """

    st.markdown("## Platform Dones")
    # Displays the Table of platform dones
    st.dataframe(platform_dones_table, use_container_width=use_container_width)
    _counts_df = (
        platform_dones_table.groupby(["platform_name", "done_name"], as_index=False)
        .agg({"done_triggered": sum})
        .rename(columns={"done_triggered": "count"})
    )
    x_value = "count"
    if st.checkbox("Show as Percentage[%]", key="platform_dones_percentage"):
        _counts_df["percentage"] = convert_counts_to_percentage(input_table=_counts_df, count_column="count")
        x_value = "percentage"
    fig = px.bar(_counts_df, x=x_value, y="done_name", facet_col="platform_name", facet_col_wrap=2)
    st.plotly_chart(fig, use_container_width=use_container_width)


def agent_dones_counts_and_table(
    agent_dones_table: pd.DataFrame, use_container_width: bool, sess_state: SessionStateProxy
) -> SessionStateProxy:
    """Generates a streamlit table of agent dones (data in `agent_dones_table`)
    and barcharts of WLD for a given done condition. The done conditions are selected
    by the user in a drop-down menu.

    Parameters
    ----------
    agent_dones_table : pd.DataFrame
        The agent dones table from EpisodeArtifactTables
    use_container_width : bool
        When True, uses the full container width (streamlit parameter)
    sess_state : SessionStateProxy
        The streamlit session state.

    Returns
    -------
    SessionStateProxy
        The streamlit session state.
    """

    st.markdown("## Agent Dones")
    st.dataframe(agent_dones_table, use_container_width=use_container_width)
    # Get a list of the agent dones
    dones = agent_dones_table["done_name"].unique().tolist()
    dones.sort()
    done_statuses = agent_dones_table["done_status"].unique().tolist()
    # Creates a select box for the done names
    selected_done = st.selectbox(label="Done Name", options=dones)

    # Creates a section of check boxes for the user to select which done
    # statuses (Win, Loss, Draws etc.) should be included
    st.markdown("**Select Done Statuses to Include**")
    done_status_selections = {}
    columns = st.columns(6)
    for idx, col in enumerate(done_statuses):
        with columns[idx]:
            default_value = False
            if col in ["WIN", "PARTIAL_WIN"]:
                default_value = True
            done_status_selections[col] = st.checkbox(label=col, value=default_value)
    include_status = [done_status for done_status, selected in done_status_selections.items() if selected]
    # User option to show as a percentage
    as_percentage = st.checkbox("Show as Percentage[%]")
    if include_status or sess_state.get("agent_done_status_plot_generated"):
        sess_state.agent_done_status_plot_generated = True
        _idx_done_name_filter = agent_dones_table["done_name"] == selected_done
        _idx_selected_statuses = agent_dones_table["done_status"].isin(include_status)
        _subset_df = agent_dones_table.loc[_idx_done_name_filter & _idx_selected_statuses]
        _counts_df = (
            _subset_df.groupby(["agent_id", "done_status"], as_index=False)
            .agg({"test_case": "count"})
            .rename(columns={"test_case": "count"})
        )
        x_value = "count"
        if as_percentage:
            _counts_df["percentage"] = convert_counts_to_percentage(input_table=_counts_df, count_column="count")
            x_value = "percentage"
        fig = px.bar(_counts_df, x=x_value, y="agent_id", color="done_status")
        st.plotly_chart(fig, use_container_width=use_container_width)

    return sess_state



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from collections import OrderedDict

import gymnasium

from corl.dones.done_func_base import DoneFuncBase, DoneFuncBaseValidator, DoneStatusCodes
from corl.libraries.units import Quantity
from corl.simulators.base_simulator import BaseSimulatorState


class EpisodeLengthDoneValidator(DoneFuncBaseValidator):
    """Initialize an EpisodeLengthDone

    Parameters
    ----------
    horizon : float, optional
        The max expected length for horizon (in seconds), by default 1000
    """

    horizon: Quantity


class EpisodeLengthDone(DoneFuncBase):
    """
    CheckEpisodeLengthDone Just checks to see if we hit the timeout time and notes in the code
    via a done condition... Note this is largely a debug item
    """

    REQUIRED_UNITS = {"horizon": "second"}

    def __init__(self, **kwargs) -> None:
        self.config: EpisodeLengthDoneValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[EpisodeLengthDoneValidator]:
        """Returns the validator for this done condition"""
        return EpisodeLengthDoneValidator

    def __call__(
        self,
        observation: OrderedDict,
        action: OrderedDict,
        next_observation: OrderedDict,
        next_state: BaseSimulatorState,
        observation_space: gymnasium.Space,
        observation_units: OrderedDict,
    ) -> bool:
        try:
            done = next_state.sim_time >= self.config.horizon.m

        except ValueError:
            # Missing platform should trigger some other done condition
            done = False

        if done:
            next_state.episode_state[self.platform][self.name] = DoneStatusCodes.DRAW

        return done



"""
This module defines functions that determine terminal conditions for the 1D Docking environment.
"""

from corl.dones.done_func_base import DoneFuncBase, DoneStatusCodes
from corl.simulators.common_platform_utils import get_platform_by_name
from corl.simulators.pong.paddle_platform import PaddleType
from corl.simulators.pong.pong import GameStatus


class PongGameDoneFunction(DoneFuncBase):
    """
    A done function that determines if the game status is no longer in progress and
    determines if the paddle this done is assigned to won or lost
    """

    def __call__(self, observation, action, next_observation, next_state, observation_space, observation_units):
        """
        Parameters
        ----------
        observation : np.ndarray
            np.ndarray describing the current observation
        action : np.ndarray
            np.ndarray describing the current action
        next_observation : np.ndarray
            np.ndarray describing the incoming observation
        next_state : np.ndarray
            np.ndarray describing the incoming state

        Returns
        -------
        done : DoneDict
            dictionary containing the condition condition for the current agent

        """

        paddle_type = get_platform_by_name(next_state, self.config.platform_name).paddle_type
        game_status = next_state.game_status

        done = False
        if game_status is not GameStatus.IN_PROGRESS:
            done = True
            if self._win_left(game_status, paddle_type) or self._win_right(game_status, paddle_type):
                next_state.episode_state[self.config.platform_name][self.name] = DoneStatusCodes.WIN
            else:
                next_state.episode_state[self.config.platform_name][self.name] = DoneStatusCodes.LOSE
        # else:
        #     next_state.episode_state[self.config.platform_name][self.name] = DoneStatusCodes.DRAW

        return done

    @staticmethod
    def _win_left(game_status: GameStatus, paddle_type: PaddleType):
        """Check if the paddle_type is left and left one the game

        Parameters
        ----------
        game_status : GameStatus
            status of current game
        paddle_type : PaddleType
            paddle type of the platform being check

        Returns
        -------
        done : bool
            return true if the paddle type one the game
        """

        return game_status == GameStatus.LEFT_WIN and paddle_type is PaddleType.LEFT

    @staticmethod
    def _win_right(game_status: GameStatus, paddle_type: PaddleType):
        """Check if the paddle_type is right and right one the game

        Parameters
        ----------
        game_status : GameStatus
            status of current game
        paddle_type : PaddleType
            paddle type of the platform being check

        Returns
        -------
        done : bool
            return true if the paddle type one the game
        """
        return game_status == GameStatus.RIGHT_WIN and paddle_type is PaddleType.RIGHT



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import typing

from corl.evaluation.metrics.metric import Metric, NonTerminalMetric
from corl.evaluation.metrics.types.nonterminals.timed_value import TimedValue


class Vector(NonTerminalMetric):
    """NonTerminal Metric to contain a vector of other metrics

    This class can be inherited extend other vectorized data, like TimeSeries
    """

    _arr: list[Metric]
    _is_time_series: bool

    def __init__(self, arr: list[Metric]):
        self._arr = arr

        # Not sure what a "generator" is
        self._is_time_series = all(isinstance(item, TimedValue) for item in self._arr)

    @property
    def arr(self) -> list[Metric]:
        """Get list of metrics stored by the vector"""
        return self._arr

    def is_time_series(self) -> bool:
        """Determine if the vector is a time series"""
        return self._is_time_series

    def __add__(self, rhs: typing.Any):
        raise NotImplementedError()

    def __truediv__(self, rhs: typing.Any):
        raise NotImplementedError()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Abstraction Class for 6DOF platform types that can provide
some properties for common usage
"""
import abc

import corl.simulators.six_dof.base_six_dof_properties as six_dof_props
from corl.libraries.units import Quantity
from corl.simulators.base_platform import BasePlatform


class Base6DOFPlatform(BasePlatform):
    """
    Further abstraction of BasePlatform, this adds some common
    6dof properties that may be useful for dealing with simulators
    using aircraft
    """

    @property
    @abc.abstractmethod
    def lat_lon(self) -> Quantity:
        """
        The position of the object in space. Details on the position format are provided by position_properties

        Returns
        -------
        np.ndarray
            The position of the object in space
        """

    position_properties = six_dof_props.LatLonProp()

    @property
    @abc.abstractmethod
    def altitude(self) -> Quantity:
        """
        The position of the object in space. Details on the position format are provided by position_properties

        Returns
        -------
        np.ndarray
            The position of the object in space
        """

    altitude_properties = six_dof_props.AltitudePropMeters()

    @property
    @abc.abstractmethod
    def orientation(self) -> Quantity:
        """
        The orientation of the platform. For orientation formatting see orientation_properties

        Returns
        -------
        np.ndarray
            The orientation of the platform
        """

    orientation_properties = six_dof_props.OrientationProp()

    @property
    @abc.abstractmethod
    def velocity_ned(self) -> Quantity:
        """get the velocity in true airspeed NED (m/s)

        Returns:
            np.ndarray -- The velocity in true airspeed NED (m/s)
        """

    velocity_ned_properties = six_dof_props.VelocityNEDProp()
    angular_velocity_properties = six_dof_props.VelocityNEDProp(
        description="angular velocity for yaw rate, pitch rate and roll rate respectively"
    )

    @property
    @abc.abstractmethod
    def acceleration_ned(self) -> Quantity:
        """gets the acceleration in the NED

        Returns:
            np.ndarray -- The acceleration in the NED
        """

    acceleration_ned_properties = six_dof_props.AccelerationNEDProp()

    @property
    @abc.abstractmethod
    def speed(self) -> Quantity:
        """Get the speed of the platform

        Returns:
            np.ndarray -- The true airspeed of the platform in m/s
        """

    # TODO FIX THIS
    speed_properties = six_dof_props.TrueAirSpeedProp(name="speed", high=[1700.0], unit="mpstas", description="true airspeed in m/s")



"""
-------------------------------------------------------------------------------
The Autonomous Capabilities Team (ACT3) Deep Reinforcement Learning (D-RL) Environment

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
-------------------------------------------------------------------------------

A set of standard config updates that can be used.
"""
from corl.evaluation.runners.section_factories.plugins.config_updater import ConfigUpdate


class DoNothingConfigUpdate(ConfigUpdate):
    """
    A config update that does nothing. Used as a placeholder/default when no config update is required.
    """

    def update(self, config):  # noqa: PLR6301
        return config



"""
This module defines functions that determine terminal conditions for the 1D Docking environment.
"""

import numpy as np

from corl.dones.done_func_base import DoneFuncBase, DoneFuncBaseValidator, DoneStatusCodes
from corl.libraries.units import Quantity
from corl.simulators.common_platform_utils import get_platform_by_name, get_sensor_by_name


class DockingDoneValidator(DoneFuncBaseValidator):
    """
    This class validates that the config contains the docking region and crash velocity data needed for
    computations in the SuccessfulDockingDoneFunction.
    """

    docking_region_radius: Quantity
    velocity_threshold: Quantity
    position_sensor_name: str
    velocity_sensor_name: str


class DockingDoneFunction(DoneFuncBase):
    """
    A done function that determines if deputy has successfully docked with the chief or not.
    """

    REQUIRED_UNITS = {"docking_region_radius": "meter", "velocity_threshold": "meter / second"}

    def __init__(self, **kwargs) -> None:
        self.config: DockingDoneValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[DockingDoneValidator]:
        """
        Parameters
        ----------
        cls : constructor function

        Returns
        -------
        SuccessfulDockingDoneValidator
            config validator for the SuccessfulDockingDoneFunction

        """
        return DockingDoneValidator

    def __call__(self, observation, action, next_observation, next_state, observation_space, observation_units):
        """
        Parameters
        ----------
        observation : np.ndarray
            np.ndarray describing the current observation
        action : np.ndarray
            np.ndarray describing the current action
        next_observation : np.ndarray
            np.ndarray describing the incoming observation
        next_state : np.ndarray
            np.ndarray describing the incoming state

        Returns
        -------
        done : DoneDict
            dictionary containing the condition condition for the current agent

        """
        deputy = get_platform_by_name(next_state, self.config.platform_name)

        position_sensor = get_sensor_by_name(deputy, self.config.position_sensor_name)
        velocity_sensor = get_sensor_by_name(deputy, self.config.velocity_sensor_name)

        position = position_sensor.get_measurement()
        velocity = velocity_sensor.get_measurement()

        chief_position = np.array([0])
        docking_region_radius = self.config.docking_region_radius

        distance = abs(position.m - chief_position).item()
        in_docking = distance <= docking_region_radius.m

        max_velocity_exceeded = self.config.velocity_threshold.m < abs(velocity.m).item()

        successful_dock = in_docking and not max_velocity_exceeded
        crash = in_docking and max_velocity_exceeded

        done = False
        if crash:
            done = True
            next_state.episode_state[self.platform][self.name] = DoneStatusCodes.LOSE
        elif successful_dock:
            done = True
            next_state.episode_state[self.platform][self.name] = DoneStatusCodes.WIN

        return done



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Configures teams to perform in evaluation
"""
import typing
from typing import Annotated

from pydantic import BaseModel, BeforeValidator, ConfigDict

import corl.evaluation.loader.i_agent_loader as agent_loader
from corl.evaluation.evaluation_factory import FactoryLoader
from corl.parsers.agent_and_platform import CorlAgentConfigArgs, CorlPlatformConfigArgs


class LoadableCorlAgent(CorlAgentConfigArgs):
    """An Agent on a platform"""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    agent_loader: Annotated[agent_loader.IAgentLoader, BeforeValidator(FactoryLoader.resolve_factory)]


class Teams(BaseModel):
    """Describes teams that are to participate in an evaluation"""

    platform_config: list[CorlPlatformConfigArgs]
    agent_config: list[LoadableCorlAgent]

    def iterate_on_participant(self, func: typing.Callable[[LoadableCorlAgent], typing.Any]) -> list[typing.Any]:
        """Perform some action on each particpant and aggregate results to an array

        Arguments:
            func {typing.Callable[Agent]} -- function to perform

        Returns:
            [type] -- Aggregated array from each particpant on each team.
        """
        return [func(agent) for agent in self.agent_config]



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Generate test cases from configuration
"""
import itertools
import logging

import pandas as pd
from numpy.random import BitGenerator, Generator

from corl.evaluation.eval_logger_name import EVAL_LOGGER_NAME
from corl.evaluation.runners.section_factories.test_cases import config_parser


def create_test_cases(config: config_parser.RESOLVED_VARIABLE_STRUCTURE_VALUE) -> pd.DataFrame:
    """Create the Cartesian product of test cases from a configuration.

    Parameters
    ----------
    config : config_parser.RESOLVED_VARIABLE_STRUCTURE_VALUE
        Single level configuration mapping as if produced by `config_parser.resolve_config_ranges`.  The keys may contain "." to represent
        nesting levels.  Each value is a sequence of possible values for that variable.

    Returns
    -------
    pandas.DataFrame
        Collection of all test cases from the initial condition configuration.  The columns are the variables from the keys of the input
        dictionary.  Each row is a unique test case.
    """
    logger = logging.getLogger(EVAL_LOGGER_NAME)
    num_cases = 1
    for v in config.values():
        num_cases *= len(v)
    if num_cases > 1000000:
        logger.warning("Number of test cases: %s", num_cases)

    keys = sorted(config.keys())
    values = [config[key] for key in keys]

    return pd.DataFrame(itertools.product(*values), columns=keys)


def sample(bgen: BitGenerator, test_cases: pd.DataFrame, amount: int | float) -> pd.DataFrame:
    """Sample test cases by count or percentage.

    Parameters
    ----------
    bgen : numpy.random.BitGenerator
        Underlying random number generator
    test_cases : pandas.DataFrame
        Test cases to sample.  Rows are test cases.  Columns are variables.
    amount : Union[int, float]
        Count (if >= 1) or percentage (if < 1) of test cases to return

    Returns
    -------
    pandas.DataFrame
        Sampled test cases, with fewer rows according to amount provided.
    """

    return test_cases.sample(n=int(amount), random_state=bgen) if amount >= 1 else test_cases.sample(frac=amount, random_state=bgen)


def generate_random(gen: Generator, config: config_parser.RANDOMIZE, value: pd.Series, variable: str) -> pd.Series:
    """Generates random samples from the distribution parameters

    Parameters
    ----------
    gen : numpy.random.Generator
        Underlying random number generator
    config : Mapping[str, Union[str, float]]
        Configuration for the distribution.  The name of the distribution is the key `distribution`.  The remaining keys are dependent on
        the type of distribution.  Supported distributions with their parameters are:
          - uniform: Uniform distribution
              - width: The width of the distribution, centered at value
          - angular_uniform: Uniform distribution on a periodic domain, such as angles.
              - width: The width of the distribution, ignoring the branch cut
              - minimum: Branch cut location at the minimum edge of the periodicity.  For example, -180.
              - maximum: Branch cut location at the maximum edge of the periodicity.  For example, 180.
          - normal: Normal distribution
              - scale: The standard deviation
    value : pd.Series
        Current value of the variable about which to randomize.
    variable : str
        Name of the variable for error messages

    Returns
    -------
    value : pd.Series
        Updated value

    Raises
    ------
    ValueError
        Improper parameters for the selected distribution
        Unknown distribution
    """

    value = value.copy()

    distribution = str(config["distribution"]).lower()

    if distribution == "uniform":
        if set(config.keys()) != {"distribution", "width"}:
            raise ValueError(f"{variable} does not contain proper parameters for uniform distribution")

        extent = float(config["width"]) / 2
        value += gen.uniform(low=-extent, high=extent, size=len(value))

    elif distribution == "angular_uniform":
        if set(config.keys()) != {"distribution", "width", "min", "max"}:
            raise ValueError(f"{variable} does not contain proper parameters for angular uniform distribution")

        extent = float(config["width"]) / 2
        value += gen.uniform(low=-extent, high=extent, size=len(value))

        max_val = float(config["max"])
        min_val = float(config["min"])
        period = max_val - min_val

        while any(value > max_val):
            value.loc[value > max_val] -= period
        while any(value < min_val):
            value.loc[value < min_val] += period

    elif distribution == "normal":
        if set(config.keys()) != {"distribution", "scale"}:
            raise ValueError(f"{variable} does not contain proper parameters for normal distribution")

        value += gen.normal(loc=0, scale=float(config["scale"]), size=len(value))

    else:
        raise ValueError(f'{variable} contains unknown distribution {config["distribution"]}')

    return value


def randomize_test_cases(
    test_cases: pd.DataFrame, gen: Generator, config: config_parser.EXTRACTED_VARIABLE_STRUCTURE_RANDOMIZE
) -> pd.DataFrame:
    """Randomize the values contained within a collection of test cases.

    Parameters
    ----------
    test_cases : pandas.DataFrame
        Test cases over which to randomize.  Columns are variables.  Rows are samples.  This value is updated in-place.
    gen : numpy.random.Generator
        Underlying random number generator
    config : config_parser.EXTRACTED_VARIABLE_STRUCTURE_RANDOMIZE
        Arguments for the distribution of each variable.  The keys of the outer mapping are the variable names, all of which must be columns
        in test_cases.  The inner mapping are the arguments for the random distribution.  See `generate_random` for the format of the inner
        mapping.

    Returns
    -------
    pd.DataFrame
        Updated test cases.

    Raises
    ------
    ValueError
        Variable names in configuration is not present in the test case columns.
    """

    test_cases = test_cases.copy()

    if unknown_columns := set(config.keys()) - set(test_cases.columns):
        raise ValueError(f"Unknown columns: {unknown_columns}")

    for col, col_config in config.items():
        test_cases.loc[:, col] = test_cases.loc[:, col].astype(float)
        test_cases.loc[:, col] = generate_random(gen=gen, config=col_config, value=test_cases.loc[:, col], variable=col)

    return test_cases



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Contains classes to represent
"""
import dataclasses
from collections import defaultdict
from typing import Any

import pandas as pd
from pydantic import validator

from corl.evaluation.episode_artifact import EpisodeArtifact


@dataclasses.dataclass
class EvaluationOutcome:
    """Dataclass to hold the results of an evaluation run"""

    test_cases: pd.DataFrame | list[dict[str, Any]]
    episode_artifacts: dict[int, list[EpisodeArtifact]] = None  # type: ignore

    @validator("episode_artifacts", pre=True, always=True)
    def gen_default_dict(cls, v):
        if v is None:
            return defaultdict(list)
        return v

    def get_test_cases(self) -> pd.DataFrame:
        if isinstance(self.test_cases, pd.DataFrame):
            return self.test_cases
        return pd.DataFrame(self.test_cases)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from abc import abstractmethod
from collections.abc import Callable
from typing import Any, Generic, TypeVar, final
from weakref import CallableProxyType, ReferenceType, proxy, ref

T = TypeVar("T")


# class _TypeChecker:
#     def __init__(self, type_: tuple[type, ...] | Any = ()):
#         self.type_ = type_

#     def is_base_of(self, type_: "type | (_TypeChecker | Any)") -> bool:
#         if self.type_ == Any:
#             return True

#         if type_ == Any:
#             return False

#         if isinstance(type_, _TypeChecker):
#             if type_.type_ == Any:
#                 return True

#             return any(self.is_base_of(t) for t in type_.type_)

#         return issubclass(type_, self.type_)

#     def __str__(self) -> str:
#         if self.type_ == Any:
#             return f"[{Any.__qualname__}]"
#         return "[" + " | ".join(f"{t.__qualname__}" for t in self.type_) + "]"


# class _GenericType(Generic[T]):
#     @lru_cache
#     def get_generic_type(self, target: type) -> _TypeChecker:
#         for orig_base in self.__orig_bases__:
#             orig_type = get_origin(orig_base)
#             if orig_type is None:
#                 orig_type = orig_base
#             if issubclass(orig_type, target):
#                 types = get_args(orig_base)
#                 assert len(types) == 1, "There should only be one 'template' type."
#                 type_ = types[0]
#                 if type_ is None or isinstance(type_, TypeVar):
#                     return _TypeChecker(Any)

#                 arg_types = get_args(type_)
#                 if arg_types:
#                     return _TypeChecker(arg_types)

#                 return _TypeChecker((type_,))

#         return _TypeChecker()


class SlotExpired(RuntimeError):
    ...


class BaseSlot(Generic[T]):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._expired = False
        self._signals: list[ReferenceType[Signal[T]]] = []  # type: ignore

    def add_signal(self, signal: "ReferenceType[Signal[T]]"):
        if signal not in self._signals:
            self._signals.append(signal)

    def get_signals(self) -> list["ReferenceType[Signal[T]]"]:
        return self._signals

    def disconnect(self):
        self._expired = True

    def on_message(self, data: T) -> Any:
        if self._expired:
            raise SlotExpired
        return self._on_message(data)

    @abstractmethod
    def _on_message(self, data: T) -> Any:
        ...

    def __setstate__(self, state):
        self.__dict__.update(state)
        for signal_ref in self._signals:
            if (signal := signal_ref()) is not None:
                signal.register(self)


@final
class Slot(BaseSlot[T]):
    def __init__(self, callback: Callable[[T], Any], *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._callback: Callable[[T], Any] = callback

    def _on_message(self, data: T) -> Any:
        return self._callback(data)


@final
class WeakSlot(BaseSlot[T]):
    def __init__(self, callback: Callable[[T], Any], *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._callback: CallableProxyType[Callable[[T], Any]] = proxy(callback)

    def _on_message(self, data: T) -> Any:
        try:
            return self._callback(data)
        except ReferenceError as err:
            raise SlotExpired from err


@final
class Signal(Generic[T]):
    def __init__(self, **kwargs):
        super().__init__()
        self._slots: list[BaseSlot[T]] = []  # type: ignore

    def register(self, slot: BaseSlot[T]) -> BaseSlot[T]:
        # signal_type = self.get_generic_type(Signal)
        # slot_type = slot.get_generic_type(Slot)

        # if not slot_type.is_base_of(signal_type):
        #     raise RuntimeError(
        #         f"Mismatched types: {type(self).__name__}{signal_type} is not able to "
        #         f"register slot {type(slot).__name__}{slot_type}"
        #     )

        ## Maintain self._slots insertion order and do not allow duplicates
        if slot not in self._slots:
            self._slots.append(slot)

            slot.add_signal(ref(self))

        return slot

    def __call__(self, data: T) -> list[Any]:
        results: list[Any] = []

        to_remove: list[BaseSlot[T]] = []
        for slot in self._slots:
            try:
                results.append(slot.on_message(data))

            except SlotExpired:  # noqa: PERF203
                to_remove.append(slot)

        for slot in to_remove:
            self._slots.remove(slot)

        return results

    def __getstate__(self):
        slots = self._slots

        self._slots = []
        state = self.__dict__.copy()
        state["slots"] = slots.copy()

        self._slots = slots

        return state



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

from corl.evaluation.metrics.generator import MetricGeneratorAggregator
from corl.evaluation.metrics.metric import Metric, TerminalMetric
from corl.evaluation.metrics.types.nonterminals.vector import Vector


class Accumulate(MetricGeneratorAggregator):
    """Aggregate by accumulating."""

    def generate_metric(self, params: list[Metric] | Metric, **kwargs) -> Metric:  # noqa: PLR6301
        """Generate the metric

        Arguments:
            params {typing.Union[typing.List[Metric], Metric]} -- Already computed metrics

        Returns:
            Metric -- Computed Metric
        """

        arr: list[Metric]
        if isinstance(params, Metric):
            arr = [params]
        elif isinstance(params, list):
            arr = []
            for idx, item in enumerate(params):
                if not isinstance(item, Metric):
                    raise RuntimeError(f'{idx}th item in list of metrics to accumulate is not of type "Metric"')

                # If it's a terminal metric then just merge
                if isinstance(item, TerminalMetric):
                    arr.append(item)
                elif isinstance(item, Vector):
                    arr.extend(iter(item.arr))
                else:
                    raise RuntimeError(f"Unsure how to merge {type(item)}")
        else:
            raise RuntimeError("Expecting either a singular metric or a list of Metrics")

        return Vector(arr)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Library to extract the value function from within callbacks
"""

from collections import defaultdict

from ray.rllib import BaseEnv
from ray.rllib.evaluation.episode_v2 import EpisodeV2
from ray.rllib.policy import Policy
from ray.rllib.utils.typing import PolicyID


def extract_value_function_step_callback(
    *,
    worker,
    base_env: BaseEnv,
    policies: dict[PolicyID, Policy] | None = None,
    episode: EpisodeV2,
    **kwargs,
) -> dict[str, float]:
    """Extract the value function from information provided by on_episode_step"""

    if policies is None:
        return {}

    env = base_env.get_sub_environments()[episode.env_id]

    non_trainable_agent_dict = {k: v for (k, v) in env.agent_dict.items() if not (v.trainable)}
    # Map of policy_id to agent_id(s)
    policy_id_map = defaultdict(list)
    for agent_id in env.agent_to_platforms:
        if agent_id in non_trainable_agent_dict:
            continue
        t_policy_id = episode.policy_for(agent_id)
        policy_id_map[t_policy_id].append(agent_id)

    # Pull out state value for each agent from policy.model.value_function
    output_dict = {}
    for agent_id in env.agent_to_platforms:
        if agent_id in non_trainable_agent_dict:
            continue
        policy_id = episode.policy_for(agent_id)
        policy = policies[policy_id]

        # Tensor size is the num obs fed into the vf in a forward pass
        try:
            vf_tensor = policy.model.value_function().tolist()  # type: ignore
        except AttributeError:
            continue

        # Assume first value in tensor
        vf_value = vf_tensor[0]

        # Determine tensor values for agents sharing a policy
        if len(policy_id_map[policy_id]) > 1:
            # Pull out the set of observations used for the previous VF model forward pass
            # https://docs.ray.io/en/latest/rllib/package_ref/env/base_env.html#ray.rllib.env.base_env.BaseEnv.poll
            # (1) New observations for each ready agent
            # (2) Reward values for each ready agent. If the episode is just started, the value will be None.
            # (3) Terminated values for each ready agent. The special key “__all__” is used to indicate episode termination.
            # (4) Truncated values for each ready agent. The special key “__all__” is used to indicate episode truncation.
            # (5) Info values for each ready agent. Agents may take off-policy actions, in which case, there will be an
            #     entry in this dict that contains the taken action.
            # (6) There is no need to send_actions() for agents that have already chosen off-policy actions.
            trainable_observations, _, agent_termination_values, _, _, _ = base_env.poll()

            # Indicates episode ended, no way to identify owner of vf_value
            # Indicates agent's obs not provided in previous model forward pass - no value
            if len(trainable_observations) == 0 or agent_id not in trainable_observations[episode.env_id]:
                vf_value = 0
            else:
                # Build agent_id list for a shared policy
                shared_policy_agent_ids = [
                    agent_name
                    for agent_name in trainable_observations[episode.env_id]
                    if agent_name in policy_id_map[policy_id] and not bool(agent_termination_values[episode.env_id].get(agent_name))
                ]
                # Determine index into vf_tensor given agent_id in shared_policy list
                if agent_id in shared_policy_agent_ids:
                    index = shared_policy_agent_ids.index(agent_id)
                    vf_value = vf_tensor[index]

        output_dict[agent_id] = vf_value

    return output_dict



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from pathlib import Path
from typing import Annotated, Any

from pydantic import BaseModel, BeforeValidator

from corl.parsers.yaml_loader import load_file


def try_load_file(value: dict[str, Any] | (str | Path)) -> dict[str, Any]:
    if isinstance(value, str | Path):
        return load_file(value)
    return value


class CorlPlatformConfigArgs(BaseModel):
    """
    name: name of the platform
    config: path to the config file this platform should be configured with
    """

    name: str
    config: Annotated[dict[str, Any], BeforeValidator(try_load_file)]


class CorlPlatformsConfigArgs(BaseModel):
    """
    platform_config: list of platforms
    """

    platform_config: list[CorlPlatformConfigArgs]


class CorlAgentConfigArgs(BaseModel):
    """
    name: name of the agent
    platforms: a list of platforms the agent at least partially controls
    config: path to the agent config for this agent
    policy: path to the policy config for this agent
    """

    name: str
    platforms: list[str]
    config: Annotated[dict[str, Any], BeforeValidator(try_load_file)]
    policy: Annotated[dict[str, Any], BeforeValidator(try_load_file)]



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

AvailablePlatforms
"""
import logging
import typing
from collections import OrderedDict
from functools import cached_property

import numpy as np
from gymnasium.spaces import Dict
from pydantic import field_validator

from corl.glues.base_glue import BaseAgentControllerGlue
from corl.glues.base_wrapper import BaseWrapperGlue, BaseWrapperGlueValidator
from corl.libraries.env_space_util import EnvSpaceUtil
from corl.libraries.property import DictProp
from corl.libraries.units import corl_get_ureg


class DeltaActionValidator(BaseWrapperGlueValidator):
    """
    step_size:      A dict that contains a floating point scalar for each action in the action space,
                    by which the corresponding delta action is scaled prior to converting the action
                    to the wrapped space.
                    e.g. A throttle DeltaAction.apply_action([0.2]) with step_size=[.05] would move the
                    absolute throttle position to 0.01 higher than it was at the end of the last step.
    """

    step_size: float = 1.0
    is_wrap: bool

    @field_validator("step_size")
    @classmethod
    def check_step_scale(cls, v):
        """
        verifies range of step scale values
        """
        if v >= 1.0 or v < 0:
            raise ValueError("DeltaActionValidator got step size of more that 1.0 or less than 0")
        return v


class DeltaAction(BaseWrapperGlue):
    """
    DeltaAction is a glue class that wraps another glue class.
    It treats the actions passed to it as a delta from the last action command rather
    E.G. if the wrapped action space has has throttle as one of the controls, then a delta action of
    0.2 would move the absolute throttle position 0.2 higher than it was at the end of the last step.
    """

    def __init__(self, **kwargs) -> None:
        self.config: DeltaActionValidator
        super().__init__(**kwargs)

        self._logger = logging.getLogger(DeltaAction.__name__)

        wrapped_aspace = self.glue().action_space
        assert wrapped_aspace is not None
        self.step_size = EnvSpaceUtil.convert_config_param_to_space(action_space=wrapped_aspace, parameter=self.config.step_size)

        self._is_wrap = self.config.is_wrap

        self.saved_action_deltas: EnvSpaceUtil.sample_type = OrderedDict()
        assert isinstance(self.action_space, Dict)
        for space_name in self.action_space:
            self.saved_action_deltas[space_name] = corl_get_ureg().Quantity(
                np.array([0], dtype=self.action_prop.spaces[space_name].dtype), self.action_prop.spaces[space_name].get_units()
            )

        inner_glue: BaseAgentControllerGlue = typing.cast(BaseAgentControllerGlue, self.glue())
        if not isinstance(inner_glue, BaseAgentControllerGlue):
            raise TypeError(f"Inner glue is not a BaseAgentControllerGlue, but rather {type(inner_glue).__name__}")

        wrapped_glue_name = self.glue().get_unique_name()
        self._uname = None if wrapped_glue_name is None else f"{wrapped_glue_name}Delta"

    @staticmethod
    def get_validator() -> type[DeltaActionValidator]:
        """Return validator"""
        return DeltaActionValidator

    def get_unique_name(self) -> str:
        """Class method that retrieves the unique name for the glue instance"""
        return self._uname

    def get_observation(self, other_obs: OrderedDict, obs_space, obs_units) -> EnvSpaceUtil.sample_type:
        """Get observation"""
        return {"absolute": self.glue().get_observation(other_obs, obs_space, obs_units), "delta": self.saved_action_deltas}

    @cached_property
    def observation_prop(self):
        return DictProp(spaces={"absolute": self.glue().observation_prop, "delta": self.action_prop})

    @cached_property
    def action_prop(self):
        # get the action space from the parent
        original_action_space = self.glue().action_prop
        assert original_action_space is not None
        # zero mean the space so we can scale it easier
        zero_mean_space = original_action_space.zero_mean()

        # scale the size of the unbiased space
        for space_name, space in zero_mean_space.spaces.items():
            zero_mean_space.spaces[space_name] = space.scale(scale=self.step_size[space_name])

        return zero_mean_space

    def apply_action(
        self, action: EnvSpaceUtil.sample_type, observation: EnvSpaceUtil.sample_type, action_space, obs_space, obs_units
    ) -> None:
        """
        Apply the action for the controllers
        """

        self._logger.debug(f"apply_action: {action}")

        inner_glue: BaseAgentControllerGlue = typing.cast(BaseAgentControllerGlue, self.glue())
        if not isinstance(inner_glue, BaseAgentControllerGlue):
            raise TypeError(f"Inner glue is not a BaseAgentControllerGlue, but rather {type(inner_glue).__name__}")

        last_absolute_action = inner_glue.get_applied_control()

        assert self.action_space is not None
        absolute_action = EnvSpaceUtil.add_space_samples(
            space_template=self.action_space,
            space_sample1=action,
            space_sample2=last_absolute_action,
        )
        assert inner_glue.action_space is not None
        absolute_action = EnvSpaceUtil.clip_space_sample_to_space(absolute_action, inner_glue.action_space, self._is_wrap)

        self.saved_action_deltas = action

        inner_glue.apply_action(absolute_action, observation, action_space, obs_space, obs_units)

    def get_info_dict(self):  # noqa: PLR6301
        """
        Get the user specified metadata/metrics/etc.
        """
        return {}



"""
This module defines the controller used by the agent to interact with its environment.
"""


from corl.libraries.plugin_library import PluginLibrary
from corl.libraries.units import Quantity
from corl.simulators.base_parts import BaseController
from corl.simulators.docking_1d.available_platforms import Docking1dAvailablePlatformType
from corl.simulators.docking_1d.properties import ThrustProp
from corl.simulators.docking_1d.simulator import Docking1dSimulator


class Thrust1dController(BaseController):
    """
    A controller to apply thrust along a single axis.

    Parameters
    ----------
    parent_platform : Docking1dPlatform
        the platform to which the controller belongs
    config : dict
        contains configuration properties
    control_properties : corl.libraries.property.BoxProp
        a class to define the acceptable bounds and units of the controller's control
    """

    def __init__(
        self,
        parent_platform,
        config,
        control_properties=ThrustProp,
    ) -> None:
        super().__init__(property_class=control_properties, parent_platform=parent_platform, config=config)

    @property
    def name(self):
        """
        Returns
        -------
        String
            name of the controller
        """
        return self.config.name

    def apply_control(self, control):
        """
        Applies control to the parent platform

        Parameters
        ----------
        control
            ndarray describing the control to the platform
        """
        assert isinstance(control, Quantity)
        self.parent_platform.save_action_to_platform(action=control)

    def get_applied_control(self) -> Quantity:
        """
        Retrieve the applied control to the parent platform

        Returns
        -------
        Quantity
            Previously applied action

        """
        return self.parent_platform.get_applied_action()


# The defined controller must be registered with the PluginLibrary, along with a name for reference in the config, and
# a dict defining associated Simulator class and platform type enum.

PluginLibrary.AddClassToGroup(
    Thrust1dController, "1D_Controller_Thrust", {"simulator": Docking1dSimulator, "platform_type": Docking1dAvailablePlatformType}
)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Classes that represent a metric generator

A metric generator is a class that generates a metric.
"""
import abc
import typing

from pydantic import BaseModel, field_validator

from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.evaluation.metrics.metric import Metric
from corl.evaluation.metrics.scopes import Scopes, from_string

Type = typing.TypeVar("Type")


class MetricGenerator(BaseModel, typing.Generic[Type]):
    """Base class for a metric generator

    A metric generator is a class that is used to create a metric from parameters
    """

    name: str
    """Name of metric
    """
    description: str = "None provided"
    """
    Description of the metric i.e. what is it calculating
    """

    @abc.abstractmethod
    def generate_metric(self, params: Type, **kwargs) -> Metric:
        """Generate the metric from given parameters

        Arguments:
            params {Type} -- Parameters to generate the metric from

        Returns:
            Metric -- Generated metric
        """


class MetricGeneratorTerminalEventScope(abc.ABC, MetricGenerator[EpisodeArtifact]):
    """Base class for a generator that produces metrics from a single Event"""


class MetricGeneratorTerminalEvaluationScope(abc.ABC, MetricGenerator[list[EpisodeArtifact]]):
    """Base class for a generator that produces metrics from an evaluation: a set of events"""


class MetricGeneratorTerminalTournamentScope(abc.ABC, MetricGenerator[list[list[EpisodeArtifact]]]):
    """Base class for a generator that produces metrics from an tournament: a set of evaluations"""


class MetricGeneratorAggregator(abc.ABC, MetricGenerator[Metric | list[Metric]]):
    """Base class for a generator that generates it's metric by aggregating from already generated metrics"""

    metrics_to_use: str
    """Metric to use in aggregation
    """

    scope: Scopes | None = None
    """Scope that the metric may apply to to, set to None to resolve at first available
    """

    @field_validator("scope", mode="before")
    @classmethod
    def _resolve_scope(cls, v):
        return from_string(v) if v is not None and isinstance(v, str) else v



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Base Simulator and Platform for Toy OpenAI Environments
This mainly shows a "how to use example" and provide an setup to unit test with
"""
import typing
from functools import cached_property

import gymnasium
from pydantic import BaseModel, ImportString

from corl.libraries.plugin_library import PluginLibrary
from corl.libraries.units import corl_get_ureg
from corl.simulators.base_platform import BasePlatform, BasePlatformValidator
from corl.simulators.base_simulator import AgentConfig, BaseSimulator, BaseSimulatorState, BaseSimulatorValidator


class GymnasiumPlatformValidator(BasePlatformValidator):
    """GymnasiumPlatformValidator

    Parameters
    ----------
        platform: gymnasium.Env
            Gymnasium env associated with GymnasiumPlatform
    """

    platform: gymnasium.Env


class GymnasiumPlatformConfig(BaseModel):
    """
    GymnasiumPlatformSimConfig
    """

    platform_class: ImportString


class GymnasiumAgentConfig(AgentConfig):
    """
    platform_config: any configuration needed for the simulator to
                initialize this platform and configure it in the sim class
    parts_list: a list of tuples where the first element is come python class path
                    of a BasePart, and then the second element is a configuration dictionary for that part

    Arguments:
        BaseModel: [description]
    """

    platform_config: GymnasiumPlatformConfig


class GymnasiumPlatform(BasePlatform):
    """
    The GymnasiumPlatform wraps some gymnasium environment as it's platform and
    allows for saving an action to the platform for when the platform needs
    to give an action to the environment during the environment step function
    """

    def __init__(self, **kwargs) -> None:
        # hack to get this working until platforms are fixed
        self.config: GymnasiumPlatformValidator = self.get_validator()(**kwargs)
        self._platform = self.config.platform

        super().__init__(**kwargs)

        if isinstance(self.action_space, gymnasium.spaces.Discrete):
            self._last_applied_action = corl_get_ureg().Quantity(0, "dimensionless")
        elif isinstance(self.action_space, gymnasium.spaces.Box):
            self._last_applied_action = corl_get_ureg().Quantity(self.action_space.low, "dimensionless")
        else:
            raise RuntimeError("shouldn't have gotten here")

        self._operable = True

    @staticmethod
    def get_validator() -> type[GymnasiumPlatformValidator]:
        return GymnasiumPlatformValidator

    @cached_property
    def observation_space(self) -> gymnasium.spaces.Space:
        """
        Provides the observation space for a sensor to use

        Returns:
            gymnasium.Space -- the observation space of the platform gymnasium environment
        """
        return self._platform.observation_space

    @cached_property
    def action_space(self) -> gymnasium.spaces.Space:
        """
        Provides the action space for a controller to use

        Returns:
            gymnasium.Space -- the action space of the platform gymnasium environment
        """
        return self._platform.action_space

    def get_applied_action(self):
        """returns the action stored in this platform

        Returns:
            typing.Any -- any sort of stored action
        """
        return self._last_applied_action

    def save_action_to_platform(self, action):
        """
        saves an action to the platform if it matches
        the action space

        Arguments:
            action typing.Any -- The action to store in the platform

        Raises:
            RuntimeError: if the action attempted to be stored does not match
                        the environments action space
        """
        if not self.action_space.contains(action.m):
            raise RuntimeError(
                "Error: action attempting to be stored in platform does not match platforms action space\n"
                f"{action.m=} and {self.action_space=}"
            )
        self._last_applied_action = action

    @property
    def operable(self):
        return self._operable

    @operable.setter
    def operable(self, value):
        self._operable = value


class GymnasiumInclusivePartsPlatform(GymnasiumPlatform):
    """
    The GymnasiumInclusivePartsPlatform mirrors GymnasiumPlatform but without
    mutually exclusive parts
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        if isinstance(self.action_space, gymnasium.spaces.Discrete):
            self._last_applied_action = 0
        elif isinstance(self.action_space, gymnasium.spaces.Box):
            self._last_applied_action = self.action_space.low

        self._operable = True

        self._last_applied_action = corl_get_ureg().Quantity(self._last_applied_action, "dimensionless")


class GymnasiumSimulatorValidator(BaseSimulatorValidator):
    """
    Validator for GymnasiumSimulatorValidator

    gymnasium_env: the name of a gymnasium environment registered to the gym
            registry
    """

    # todo: maybe switch this to a ImportString and do a validator that it
    # implements gymnasium.core.Env
    gymnasium_env: str
    gymnasium_configs: typing.Mapping[str, bool | (float | (int | str)) | None] = {}
    seed: int = 1
    agent_configs: typing.Mapping[str, GymnasiumAgentConfig]
    wrappers: list[ImportString] = []


class SimulatorState(BaseSimulatorState):
    """
    obs: the current obs for each named gymnasium env
    rewards: the current reward for each named gymnasium env
    dones: the current done for each named gymnasium env
    info: the current info for each named gymnasium env
    """

    obs: dict[str, typing.Any] = {}
    rewards: dict[str, typing.Any] = {}
    truncated: dict[str, typing.Any] = {}
    terminated: dict[str, typing.Any] = {}
    info: dict[str, typing.Any] = {}


class GymnasiumSimulator(BaseSimulator):
    """
    Simulator backend for running openai Gymnasiums
    """

    @property
    def get_simulator_validator(self) -> type[GymnasiumSimulatorValidator]:
        """Return validator"""
        return GymnasiumSimulatorValidator

    def __init__(self, **kwargs) -> None:
        self.config: GymnasiumSimulatorValidator
        super().__init__(**kwargs)
        self._state: SimulatorState
        self.gymnasium_env_dict = {}
        for agent_name in self.config.agent_configs:
            env = gymnasium.make(self.config.gymnasium_env, **self.config.gymnasium_configs)  # type: ignore
            for wrapper_cls in self.config.wrappers:
                env = wrapper_cls(env)
            self.gymnasium_env_dict[agent_name] = env
        self.sim_platforms: dict = {}
        self._time = 0.0

    def get_platforms(self):
        """
        gets the current state of the simulation and makes the sim platforms

        Returns:
            typing.Dict[str, GymnasiumPlatform] -- the dict of openai gymnasium platforms
        """
        return {
            agent_name: self.config.agent_configs[agent_name].platform_config.platform_class(
                platform_name=agent_name,
                platform=agent_env,
                parts_list=self.config.agent_configs[agent_name].parts_list,
            )
            for agent_name, agent_env in self.gymnasium_env_dict.items()
        }

    def update_sensor_measurements(self):
        """
        Update and caches all the measurements of all the sensors on each platform
        """
        for plat in self.sim_platforms.values():
            for sensor in plat.sensors.values():
                sensor.calculate_and_cache_measurement(state=self._state)

    def reset(self, config):
        self._time = 0.0
        self.sim_platforms = self.get_platforms()
        self._state = SimulatorState(sim_platforms=self.sim_platforms, sim_time=self._time)

        for agent_name, agent_env in self.gymnasium_env_dict.items():
            self._state.obs[agent_name], self._state.info[agent_name] = agent_env.reset(seed=self.config.seed)

        self.update_sensor_measurements()
        return self._state

    def step(self, platforms_to_action):
        for agent_name, sim_platform in self.sim_platforms.items():
            if sim_platform.operable:
                tmp = self.gymnasium_env_dict[agent_name].step(sim_platform.get_applied_action().m)
                self._state.obs[agent_name] = tmp[0]
                self._state.rewards[agent_name] = tmp[1]
                self._state.terminated[agent_name] = tmp[2]
                self._state.truncated[agent_name] = tmp[3]
                self._state.info[agent_name] = tmp[4]
                if self._state.terminated[agent_name] or self._state.truncated[agent_name]:
                    sim_platform.operable = False

        self.update_sensor_measurements()
        self._time += 1
        self._state.sim_time = self._time
        return self._state

    @property
    def sim_time(self) -> float:
        return self._time

    @property
    def platforms(self):
        return self.sim_platforms

    def mark_episode_done(self, done_info, episode_state, metadata):
        pass

    def save_episode_information(self, dones, rewards, observations, observation_units):
        pass

    def render(self, state, mode="human"):
        """only render first environment"""
        agent = self.gymnasium_env_dict.keys()[0]
        self.gymnasium_env_dict[agent].render(mode)


PluginLibrary.AddClassToGroup(GymnasiumSimulator, "GymnasiumSimulator", {})

if __name__ == "__main__":
    tmp_config = {
        "gymnasium_env": "CartPole-v1",
        "agent_configs": {
            "blue0": {
                # the platform config object needs to be a Basecontroller or BaseSensor, this is just
                # to make pydantic happy
                "sim_config": {},
                "platform_config": [
                    ("corl.simulators.gymnasium.gymnasium_sensors.GymnasiumStateSensor", {}),
                    ("corl.simulators.gymnasium.gymnasium_controllers.GymnasiumMainController", {}),
                ],
            }
        },
    }

    tmp_sim = GymnasiumSimulator(**tmp_config)
    sim_state = tmp_sim.reset(None)
    print(tmp_sim.platforms["blue0"].name, tmp_sim.platforms[0].action_space)
    print(tmp_sim.platforms["blue0"].observation_space)
    step_result = tmp_sim.step({"blue0"})
    print(next(iter(tmp_sim.platforms["blue0"].sensors.values())).get_measurement())



"""
This module defines the controller used by the agent to interact with its environment.
"""

import pygame

from corl.libraries.plugin_library import PluginLibrary
from corl.libraries.property import DiscreteProp
from corl.libraries.units import Quantity, corl_get_ureg
from corl.simulators.base_parts import BaseController
from corl.simulators.pong.available_platforms import PongAvailablePlatformType
from corl.simulators.pong.paddle_platform import PaddlePlatform, PaddleType
from corl.simulators.pong.simulator import PongSimulator


class PaddleMoveProp(DiscreteProp):
    """
    paddle movement control property

    Parameters:
    name : str
        control property name
    n : int
        number of discrete controls
    unit : str
        unit of measurement for control input
    description : str
        description of control properties
    """

    name: str = "paddle_move"
    n: int = 3
    description: str = "paddle Move [UP, NO_MOVE, DOWN]"


class PaddleController(BaseController):
    """
    A controller to change the position of a paddle.

    Parameters
    ----------
    parent_platform : PaddlePlatform
        the platform to which the controller belongs
    config : dict
        contains configuration properties
    control_properties : corl.libraries.property.DiscreteProp
        a class to define the acceptable bounds and units of the controller's control
    """

    def __init__(
        self,
        parent_platform: PaddlePlatform,
        config,
        control_properties=PaddleMoveProp,
    ):
        super().__init__(property_class=control_properties, parent_platform=parent_platform, config=config)
        self._last_applied_action = corl_get_ureg().Quantity(0, "dimensionless")

    @property
    def name(self):
        """
        Returns
        -------
        String
            name of the controller
        """
        return self.config.name

    def apply_control(self, control: Quantity) -> None:  # type: ignore
        """
        Applies control to the parent platform

        Parameters
        ----------
        control
            ndarray describing the control to the platform
        """
        self._last_applied_action = control
        self.parent_platform.last_move = self.convert_action(control)

    def get_applied_control(self) -> Quantity:  # type: ignore
        """
        Retrieve the applied control to the parent platform

        Returns
        -------
        int
            Previously applied action

        """
        return self._last_applied_action

    def convert_action(self, discrete_action) -> int:
        """
        Retrieve the applied control to the parent platform

        Returns
        -------
        np.ndarray
            Previously applied action

        """
        if discrete_action.m == 1:
            return pygame.K_0  # any non action key

        if self.parent_platform.paddle_type == PaddleType.LEFT:
            return pygame.K_w if discrete_action.m == 0 else pygame.K_s
        # Right paddle
        return pygame.K_UP if discrete_action.m == 0 else pygame.K_DOWN


# The defined controller must be registered with the PluginLibrary, along with a name for reference in the config, and
# a dict defining associated Simulator class and platform type enum.

PluginLibrary.AddClassToGroup(
    PaddleController, "PaddleController", {"simulator": PongSimulator, "platform_type": PongAvailablePlatformType}
)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import dataclasses
import typing

from corl.evaluation.metrics.metric import Metric, NonTerminalMetric


@dataclasses.dataclass
class Dict(NonTerminalMetric):  # noqa: PLW1641
    """NonTerminal Metric to contain a dictionary of other metrics"""

    data: dict[str, Metric]

    def __add__(self, rhs: typing.Any):
        if isinstance(rhs, Dict):
            if self.data.keys() != rhs.data.keys():
                raise RuntimeError("Keys do not match when adding two Dict metrics")

            return Dict({key: self.data[key] + rhs.data[key] for key in self.data})
        raise RuntimeError(f"Can not add {type(rhs)} to a Dict metric")

    def __eq__(self, rhs: typing.Any):
        raise NotImplementedError()

    def __truediv__(self, denominator: typing.Any):
        if isinstance(denominator, int):
            return Dict({key: self.data[key] / denominator for key in self.data})
        raise RuntimeError(f"Can not divide a Dict metric by  {type(denominator)}")



import os, shutil

if __name__ == "__main__":
    src_pth = "config"
    target_path = "corl/config"

    if os.path.exists(src_pth):
        if os.path.exists(target_path) and os.path.isdir(target_path):
            shutil.rmtree(target_path)
        shutil.copytree(src_pth, target_path)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Entry/Launch point to the generate_metrics process/segment of the evaluation framework.
"""
import itertools
import pathlib
import typing

import jsonargparse
import pandas as pd
import ray.cloudpickle as pickle

from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.evaluation.evaluation_artifacts import EvaluationArtifact_EvaluationOutcome, EvaluationArtifact_Metrics
from corl.evaluation.evaluation_outcome import EvaluationOutcome
from corl.evaluation.metrics.scenario_alert_generators import ScenarioAlertGenerators
from corl.evaluation.metrics.scenario_metric_generators import ScenarioMetricGenerators
from corl.evaluation.scene_processors import SceneProcessors
from corl.parsers.yaml_loader import load_file


def get_args(path: str | None = None) -> tuple[jsonargparse.Namespace, jsonargparse.Namespace]:
    """
    Obtain the running arguments for the generate_metrics segment of the evaluation framework

    Parameters
    ----------
    path : str
        optional path to a yaml config containing running arguments
    Returns:
        params: jsonargparse.Namespace
            namespace object that contains parsed running arguments
    """
    parser = jsonargparse.ArgumentParser(description="metrics processing")
    parser.add_argument("--cfg", help="the path to a json/yml file containing the running arguments", action=jsonargparse.ActionConfigFile)

    parser.add_class_arguments(EvaluationArtifact_EvaluationOutcome, "artifact_evaluation_outcome", instantiate=True)
    parser.add_class_arguments(EvaluationArtifact_Metrics, "artifact_metrics", instantiate=True)

    # Link location such that if metrics location isn't given, the location of evaluation_outcome will be used
    def func(source_object):
        return source_object.location.absolute_path

    parser.link_arguments("artifact_evaluation_outcome", "artifact_metrics.location", compute_fn=func, apply_on="instantiate")

    parser.add_argument("--alerts-config", type=str, help="config defining which alerts to process")
    parser.add_argument("--metrics-config", type=str, help="config defining which metrics to process")
    parser.add_argument(
        "--raise-on-error-alert", type=bool, default=True, help="If true will raise an exception when error alert encountered"
    )

    if path is None:
        args = parser.parse_args()
        instantiated = parser.instantiate_classes(args)
    else:
        args = parser.parse_path(path)
        instantiated = parser.instantiate_classes(args)

    return args, instantiated


def main(params: dict[str, typing.Any] | jsonargparse.Namespace):
    """
    Run the generate_metrics process

    Parameters
    ----------
    params: jsonargparse.Namespace
        namespace object that contains parsed running arguments
    """

    # ####################
    # Pull the EvaluationOutcome object

    outcome: EvaluationOutcome = params["artifact_evaluation_outcome"].location.load()

    # ####################
    # Build the metric generators

    metric_config = params["metrics_config"]
    if isinstance(metric_config, str):
        metric_config = load_file(params["metrics_config"])
    scenario_metric_generators = ScenarioMetricGenerators.from_dict(metric_config)

    # ####################
    # Build the alert generators

    alerts_config = params["alerts_config"]
    if isinstance(alerts_config, str):
        alerts_config = load_file(params["alerts_config"])
    elif alerts_config is None:
        alerts_config = {}
    scenario_alert_generators = ScenarioAlertGenerators.from_dict(alerts_config, params["raise_on_error_alert"])

    # ####################
    # Generate the Metrics

    test_cases = outcome.test_cases
    # calculate the dataframes we care about
    # self.loaded_legacy_dataframes = self.outcome.legacy_dataframes()
    if isinstance(test_cases, pd.DataFrame):
        test_cases = test_cases.reset_index()

    evaluation_data: list[EpisodeArtifact] = list(itertools.chain(*list(outcome.episode_artifacts.values())))

    processed_data = SceneProcessors.from_evaluation(
        evaluation_data=evaluation_data,
        scenario_metric_generators=scenario_metric_generators,
        scenario_alert_generators=scenario_alert_generators,
    )

    # ####################
    # Write the metrics

    if isinstance(params["artifact_metrics"].location, str):
        metrics_out = pathlib.Path(params["artifact_metrics"].location).joinpath(params["artifact_metrics"].file)

        with open(metrics_out, "wb") as wfp:
            pickle.dump(processed_data, wfp)
    elif isinstance(params["artifact_metrics"].location, pathlib.Path):
        metrics_out = params["artifact_metrics"].location.joinpath(params["artifact_metrics"].file)

        with open(metrics_out, "wb") as wfp:
            pickle.dump(processed_data, wfp)
    else:
        raise RuntimeError("Unknown type given in `out` field")


def pre_main():
    """
    calls gets current args and passes them to main
    """
    _, instantiated = get_args()
    main(instantiated)


if __name__ == "__main__":
    pre_main()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Contains processors that execute on given data
"""
from __future__ import annotations

import abc
import typing

from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.evaluation.metrics.alerts import Alert, AlertGenerator
from corl.evaluation.metrics.generator import (
    MetricGenerator,
    MetricGeneratorAggregator,
    MetricGeneratorTerminalEvaluationScope,
    MetricGeneratorTerminalEventScope,
)
from corl.evaluation.metrics.metric import Metric, NonTerminalMetric
from corl.evaluation.metrics.scopes import Scopes

Data = typing.TypeVar("Data")


class Processor(abc.ABC):
    """Base class for a metric processor

    A metric processor combines data and generators to generate metrics
    """

    _metrics: dict[str, Metric]
    _alerts: list[Alert]
    _metrics_info: dict[str, str]  # collect metrics information

    def __init__(self, metrics: dict[str, Metric], metrics_info: dict[str, str], alerts: list[Alert] | None = None):
        self._metrics = metrics

        self._alerts = [] if alerts is None else alerts
        self._metrics_info = metrics_info

    @property
    def metrics(self) -> dict[str, Metric]:
        """Returns the calculated metrics"""
        return self._metrics

    @property
    def metrics_info(self) -> dict[str, str]:
        """
        Returns descriptions of all calculated metrics
        """

        return self._metrics_info

    @property
    def alerts(self) -> list[Alert]:
        """Returns the calculated alerts"""
        return self._alerts


class Event(Processor):
    """Process and Event

    An event is a single EpisodeArtifact instance
    """

    _data: EpisodeArtifact

    @property
    def data(self):
        """Get the data on the event"""
        return self._data

    @classmethod
    def process(
        cls, data: EpisodeArtifact, metric_generators: list[MetricGenerator], alert_generators: list[AlertGenerator], **kwargs
    ) -> Event:
        """Generate the metrics and alerts by giving the generators the provided data"""

        metrics_info: dict[str, str] = {}

        ####################
        ## Calculate metrics

        # Calculate any of the terminal metrics on the event
        metrics: dict[str, Metric] = {}
        for generator in metric_generators:
            if isinstance(generator, MetricGeneratorTerminalEventScope):
                if generator.name in metrics:
                    raise RuntimeError(f"Duplicate metric name {generator.name}")
                metrics[generator.name] = generator.generate_metric(data, **kwargs)
                metrics_info[generator.name] = generator.description

        # Calculate any aggregate generators
        for generator in metric_generators:
            if isinstance(generator, MetricGeneratorAggregator):
                aggregator = generator

                # If we have yet to computed the base metric needed for the aggregator then continue
                if aggregator.metrics_to_use not in metrics:
                    continue

                # If the scope is provided and its the wrong kind then skip
                if aggregator.scope is not None and aggregator.scope != Scopes.EVENT:
                    continue

                # If it's not a nonterminal metric then we can't aggregate anything at this point
                if not isinstance(metrics[aggregator.metrics_to_use], NonTerminalMetric):
                    continue

                # Compute and insert
                if generator.name in metrics:
                    raise RuntimeError(f"Duplicate metric name {generator.name}")
                metrics[generator.name] = aggregator.generate_metric(metrics[aggregator.metrics_to_use])
                metrics_info[generator.name] = generator.description

        ####################
        # Calculate alerts
        alerts: list[Alert] = []
        for alert_generator in alert_generators:
            # Determine if we can evaluate the alert
            can_evaluate = True
            if alert_generator.metric not in metrics:
                can_evaluate = False

            # If we can't evaluate alert, but we are marked to be able to, then error otherwise continue
            if can_evaluate is False:
                if alert_generator.scope == Scopes.EVENT:
                    raise RuntimeError(
                        f"Metric of {alert_generator.name} is marked as event scope, \
                        yet can not be called, check metrics are configured correctly"
                    )
                continue

            if isinstance(alert_generator, AlertGenerator):
                alerts += alert_generator.generate(metrics)

        event = Event(metrics, metrics_info, alerts)
        event._data = data  # noqa: SLF001
        return event


class Evaluation(Processor):
    """Process an Evaluation

    An evaluation is a list of events.
    """

    events: list[Event]

    def __init__(self, metrics: dict[str, Metric], metrics_info: dict[str, str], alerts: list[Alert], events: list[Event]):
        super().__init__(metrics, metrics_info, alerts)
        self.events = events

    @classmethod
    def process(
        cls, data: list[EpisodeArtifact], metric_generators: list[MetricGenerator], alert_generators: list[AlertGenerator], **kwargs
    ) -> Evaluation:
        """Generate the metrics by giving the generators the provided data"""

        ####################
        # Process each event with the given generators
        events: list[Event] = [Event.process(item, metric_generators, alert_generators, **kwargs) for item in data]

        ####################
        ## Calculate metrics

        metrics: dict[str, Metric] = {}
        metrics_info: dict[str, str] = {}

        # Calculate any of the terminal metrics on the Evaluation scope
        for generator in metric_generators:
            if isinstance(generator, MetricGeneratorTerminalEvaluationScope):
                if generator.name in metrics:
                    raise RuntimeError(f"Duplicate metric name {generator.name}")
                metrics[generator.name] = generator.generate_metric(data)
                metrics_info[generator.name] = generator.description

        # Calculate any aggregate generators
        for generator in metric_generators:
            if isinstance(generator, MetricGeneratorAggregator):
                aggregator = generator

                # If we have yet to computed the base metric needed for the aggregator then continue
                if aggregator.metrics_to_use not in events[0].metrics:
                    continue

                # If the scope is provided and its the wrong kind then skip
                if aggregator.scope is not None and aggregator.scope != Scopes.EVALUATION:
                    continue

                # If we have already computed the metric on an event then continue
                if aggregator.name in events[0].metrics:
                    continue

                # Extract the desired metric as a list
                metrics_to_give_aggregator = []
                for i, computed_metrics in enumerate(item.metrics for item in events):
                    if aggregator.metrics_to_use not in computed_metrics:
                        raise RuntimeError(f"{aggregator.metrics_to_use} not in the {i}th event of the evaluation")
                    metrics_to_give_aggregator.append(computed_metrics[aggregator.metrics_to_use])

                # Compute and insert
                if generator.name in metrics:
                    raise RuntimeError(f"Duplicate metric name {generator.name}")
                metrics[generator.name] = aggregator.generate_metric(metrics_to_give_aggregator)
                metrics_info[generator.name] = generator.description

        ####################
        # Calculate alerts
        alerts: list[Alert] = []
        for alert_generator in alert_generators:
            # Determine if we can evaluate the alert
            can_evaluate = True
            if alert_generator.metric not in metrics:
                can_evaluate = False

            # If we can't evaluate alert, but we are marked to be able to, then error otherwise continue
            if can_evaluate is False:
                if alert_generator.scope == Scopes.EVALUATION:
                    raise RuntimeError(
                        f"Metric of {alert_generator.name} is marked as event scope, yet can not be called,\
                        check metrics are configured correctly"
                    )
                continue

            if isinstance(alert_generator, AlertGenerator):
                alerts += alert_generator.generate(metrics)

        return Evaluation(metrics, metrics_info, alerts, events)


class Tournament(Processor):
    """Process an Evaluation

    An evaluation is a list of evaluations.
    """

    evaluations: list[Evaluation]

    @classmethod
    def process(
        cls,
        data: list[list[EpisodeArtifact]],
        metric_generators: list[MetricGenerator],
        alert_generators: list[AlertGenerator],
        **kwargs,
    ) -> Processor:
        """Generate the metrics by giving the generators the provided data"""

        raise RuntimeError("Not Tested")
        # # Process each evaluation with the given generators
        # for item in self.evaluations:
        #     item.process(generators)

        # self.metrics: typing.Dict[str, Metric] = {}

        # # Calculate any of the terminal metrics on the Evaluation scope
        # tournament_data_arr = self.data()
        # for generator in generators:
        #     if isinstance(generator, MetricGeneratorTerminalTournamentScope):
        #         if generator.name in self.metrics:
        #             raise RuntimeError(f"Duplicate metric name {generator.name}")
        #         self.metrics[generator.name] = generator.generate_metric(tournament_data_arr)

        # # Calculate any aggregate generators
        # for generator in generators:
        #     if isinstance(generator, MetricGeneratorAggregator):
        #         aggregator = generator

        #         # If we have yet to computed the base metric needed for the aggregator then continue
        #         if aggregator.metrics_to_use not in self.evaluations[0].metrics:
        #             continue

        #         # If the scope is provided and its the wrong kind then skip
        #         if aggregator.scope is not None and aggregator.scope != Scopes.TOURNAMENT:
        #             continue

        #         # Extract the desired metric as a list
        #         metrics_to_give_aggregator = []
        #         for i, metrics in enumerate([item.metrics for item in self.evaluations]):
        #             if aggregator.metrics_to_use not in metrics:
        #                 raise RuntimeError(f"{aggregator.metrics_to_use} not in the {i}th event of the evaluation")
        #             metrics_to_give_aggregator.append(metrics[aggregator.metrics_to_use])

        #         #Compute and insert
        #         if generator.name in self.metrics:
        #             raise RuntimeError(f"Duplicate metric name {generator.name}")
        #         self.metrics[generator.name] = aggregator.generate_metric(metrics_to_give_aggregator)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
This Module contains logic to update the rllib config appropriately for evaluation purposes.
"""
from pathlib import Path

from pydantic import BaseModel, ConfigDict

from corl.evaluation.runners.section_factories.plugins.config_updater import ConfigUpdate
from corl.evaluation.runners.section_factories.test_cases.test_case_manager import TestCaseStrategy


class RllibConfigUpdateValidator(BaseModel):
    """
    Initialize the function to use to update the rllib config.

    Parameters:
    -----------
    output_dir : str
        output directory for output dump of intermediate materials
    pandas_test_cases: pd.DataFrame
        test_cases to use when overriding EPP
    workers: int
        number of workers
    envs_per_worker: int
        number of environments per worker
    horizon: int
        value of using horizon , 0 or 1
    """

    output_dir: str | Path | None = None
    test_case_manager: TestCaseStrategy | None = None
    workers: int
    envs_per_worker: int | None = None
    explore: bool = False
    model_config = ConfigDict(arbitrary_types_allowed=True)


class RllibConfigUpdate(ConfigUpdate):
    """
    Initialize the function to use to update the rllib config.

    Parameters:
    -----------
    output_dir : typing.Optional[str]
        output directory for output dump of intermediate materials
    test_case_manager: typing.Optional[TestCaseStrategy]
        test_cases to use when overriding EPP
    workers: int
        number of workers
    envs_per_worker: int
        number of environments per worker
    """

    def __init__(self, **kwargs) -> None:
        self.config: RllibConfigUpdateValidator = self.get_validator()(**kwargs)

    @staticmethod
    def get_validator() -> type[RllibConfigUpdateValidator]:
        """
        retrieve validator and appropriately setup the config

        Returns
        -------
        RllibConfigUpdateValidator -- validator used to generate a configuration"""
        return RllibConfigUpdateValidator

    def update(self, config: dict):
        config["create_env_on_driver"] = True
        config["num_workers"] = self.config.workers
        if self.config.envs_per_worker is not None:
            config["num_envs_per_worker"] = self.config.envs_per_worker
        config["num_cpus_for_driver"] = 1
        config["num_cpus_per_worker"] = 1
        config["num_gpus_per_worker"] = 0
        config["num_gpus"] = 0
        if self.config.output_dir:
            config["env_config"]["output_path"] = str(self.config.output_dir)
        config["explore"] = self.config.explore
        config["batch_mode"] = "complete_episodes"

        # delegate EPP Override to TestCaseStrategy
        if self.config.test_case_manager is not None:
            config = self.config.test_case_manager.update_rllib_config(config)

        return config



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import copy
from abc import abstractmethod
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Generic, Literal, TypeVar

from pydantic import validator

from corl.episode_parameter_providers import EpisodeParameterProvider, EpisodeParameterProviderValidator, ParameterModel, Randomness
from corl.evaluation.connection.signal import BaseSlot, Signal
from corl.libraries.parameters import OverridableParameterWrapper

T = TypeVar("T")


@dataclass
class EppUpdate(Generic[T]):
    test_case_index: int
    update: T


class EppUpdateSignaller(Generic[T]):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.epp_update_signal: Signal[EppUpdate[T]] = Signal[EppUpdate[T]]()  # type: ignore


class NetworkedParameterProviderValidator(EpisodeParameterProviderValidator, Generic[T]):
    class Config:
        arbitrary_types_allowed = True

    registrar: EppUpdateSignaller[T]
    epp_variety: Literal["environment", "agent"]
    agent_id: str | None = None

    @validator("agent_id")
    def validate_agent_id(cls, v, values):
        if values["epp_variety"] == "agent" and not v:
            raise ValueError("agent_id must not be None or '' if the epp_variety is 'agent'")
        return v


class BaseNetworkParameterProvider(EpisodeParameterProvider, BaseSlot[EppUpdate[T]]):
    """EpisodeParameterProvider that receives values remotely"""

    def __init__(self, **kwargs):
        self.config: NetworkedParameterProviderValidator[T]  # type: ignore
        super().__init__(**kwargs)
        print(f"BaseNetworkParameterProvider::__init__ {id(self)}")

        self.base_params = {k: OverridableParameterWrapper(v) for k, v in self.config.parameters.items()}
        self.cur_params = copy.deepcopy(self.base_params)
        self._test_case_idx = 0

        self.config.registrar.epp_update_signal.register(self)

    @staticmethod
    def get_validator() -> type[EpisodeParameterProviderValidator]:
        return NetworkedParameterProviderValidator[T]

    def reset(self):
        self.cur_params = copy.deepcopy(self.base_params)
        self._test_case_idx = 0

    def shutdown(self):
        # del self.config.registrar
        ...

    def _do_get_params(self, rng: Randomness, env_epp_ctx: dict | None) -> tuple[ParameterModel, int | None, dict | None]:
        print(f"BaseNetworkParameterProvider::_do_get_params {id(self)} {self._test_case_idx}")
        return self.cur_params, self._test_case_idx, env_epp_ctx

    def on_message(self, epp_update: EppUpdate[T]):
        print(f"BaseNetworkParameterProvider::on_message {id(self)} {epp_update.test_case_index}")
        self.cur_params = self._do_update_params(copy.deepcopy(self.base_params), epp_update.update)
        self._test_case_idx = epp_update.test_case_index

    @abstractmethod
    def _do_update_params(
        self, params: Mapping[tuple[str, ...], OverridableParameterWrapper], updates: T
    ) -> Mapping[tuple[str, ...], OverridableParameterWrapper]:
        ...

    # def __setstate__(self, state):
    #     self.__dict__.update(state)
    #     self.config.registrar.epp_update_signal.register(Slot(self))


class TestNetworkedParameterProvider(BaseNetworkParameterProvider):
    def _do_update_params(
        self, params: Mapping[tuple[str, ...], OverridableParameterWrapper], updates  # noqa: PLR6301
    ) -> Mapping[tuple[str, ...], OverridableParameterWrapper]:
        return params



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import abc
import os
import typing

from numpy.random import Generator, RandomState
from pydantic import BaseModel, ConfigDict

from corl.libraries.parameters import Parameter, ParameterWrapper

PathLike = str | os.PathLike
Randomness = Generator | RandomState
ParameterModel = typing.Mapping[tuple[str, ...], ParameterWrapper | Parameter]


class EpisodeParameterProviderValidator(BaseModel):
    """Validation model for the inputs of EpisodeParameterProvider"""

    parameters: ParameterModel = {}
    model_config = ConfigDict(arbitrary_types_allowed=True)


class EpisodeParameterProvider(abc.ABC):
    """Interface definition for episode parameter providers."""

    def __init__(self, **kwargs) -> None:
        super().__init__()

        self.config: EpisodeParameterProviderValidator = self.get_validator()(**kwargs)

    @abc.abstractmethod
    def reset(self) -> None:
        """Resets this object to its initial state"""

    def shutdown(self) -> None:
        """Cleans up this object for destruction"""

    @staticmethod
    def get_validator() -> type[EpisodeParameterProviderValidator]:
        """Get the validator for this class."""
        return EpisodeParameterProviderValidator

    def get_params(self, rng: Randomness, env_epp_ctx: dict | None = None) -> tuple[ParameterModel, int | None, dict | None]:
        """Get the next instance of episode parameters from this provider

        Subclasses: Override _do_get_params.

        Parameters
        ----------
        rng : Union[Generator, RandomState]
            Random number generator from which to draw random values.

        Returns
        -------
        ParameterModel
            The parameters for this episode
        episode_id
            The episode index number for this set of parameters
        """

        output, episode_id, env_epp_ctx = self._do_get_params(rng, env_epp_ctx)

        if extra_keys := output.keys() - self.config.parameters.keys():
            raise KeyError(f"Extra keys provided: {extra_keys}")

        if missing_keys := self.config.parameters.keys() - output.keys():
            raise KeyError(f"Missing keys: {missing_keys}")

        if bad_types := {
            ".".join(key): type(value).__name__ for key, value in output.items() if not isinstance(value, Parameter | ParameterWrapper)
        }:
            raise TypeError(f"Unsupported` types: {bad_types}")

        return output, episode_id, env_epp_ctx

    @abc.abstractmethod
    def _do_get_params(self, rng: Randomness, env_epp_ctx: dict | None) -> tuple[ParameterModel, int | None, dict | None]:
        """Get the next instance of episode parameters from this provider

        This is an abstract method that must be overridden by each subclass.

        DO NOT CALL DIRECTLY.  USE get_params.

        Parameters
        ----------
        rng : Union[Generator, RandomState]
            Random number generator from which to draw random values.

        Returns
        -------
        ParameterCollection
            The parameters for this episode
        episode_id
            The episode index number for this set of parameters
        """
        raise NotImplementedError

    def compute_metrics(self) -> dict[str, typing.Any]:  # noqa: PLR6301
        """Get metrics on the operation of this provider.

        Often used in `on_episode_end` training callbacks.
        """
        return {}

    def update(self, results: dict, rng: Randomness) -> None:
        """Update the operation of this provider.

        Often used in `on_train_result` training callbacks.

        Parameters
        ----------
        results : dict
            As described by ray.rllib.algorithms.callbacks.DefaultCallbacks.on_train_result.
            See https://docs.ray.io/en/master/_modules/ray/rllib/agents/callbacks.html#DefaultCallbacks.on_train_result
        rng : Union[Generator, RandomState]
            Random number generator from which to draw random values.
        """

    def save_checkpoint(self, checkpoint_path: PathLike) -> None:
        """Save the internal state of the parameter provider.

        Parameters
        ----------
        checkpoint_path : PathLike
            Filesystem path at which to save the checkpoint
        """

    def load_checkpoint(self, checkpoint_path: PathLike) -> None:
        """Load the internal state from a checkpoint.

        Parameters
        ----------
        checkpoint_path : PathLike
            Filesystem path from which to restore the checkpoint
        """



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import gc
import logging
import os
import tempfile
import time
from pathlib import Path
from threading import Condition, Thread

from pydantic import BaseModel, ConfigDict
from ray.rllib.policy.sample_batch import MultiAgentBatch, SampleBatch

from corl.evaluation.connection.base_eval_connection import BaseEvalConnection, EvalConfig
from corl.evaluation.connection.signal import Slot


class EvalThreadValidator(BaseModel):
    eval_config: EvalConfig


class EvalThread(Thread):
    def __init__(self, config: dict, **kwargs):
        super().__init__(**kwargs)
        self._config = EvalThreadValidator(**config)
        self._logger = logging.getLogger(type(self).__name__)

    def run(self):
        def process_batch(batch: SampleBatch | MultiAgentBatch):
            to_process = [batch]
            for needs_processing in to_process:
                for processor in self._config.eval_config.processors:
                    try:
                        if new_result := processor(self._config.eval_config, needs_processing):
                            to_process.append(new_result)
                    except Exception as e:  # noqa: PERF203
                        self._logger.exception(e)

        # from concurrent.futures import ThreadPoolExecutor
        # with self._config.eval_schema.evaluator_config:
        #     with ThreadPoolExecutor(2) as executor:
        #         for result in self._config.evaluator(self._config.eval_schema):
        #             executor.submit(process_batch, result)

        with self._config.eval_config.experiment:
            for result in self._config.eval_config.evaluator(self._config.eval_config):
                self._logger.debug("Processing result")
                process_batch(result)

    def stop(self):
        self._config.eval_config.evaluator.stop()

    def reset(self):
        self._config.eval_config.evaluator.reset()


class EvalRunnerValidator(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    run_forever: bool = False
    tmpdir_base: Path | None = None
    connection: BaseEvalConnection | None = None


class EppReset:
    ...


class EvalRunner:
    def __init__(self, **kwargs):
        super().__init__()
        self._config = self.get_validator()(**kwargs)

        if self._config.connection is not None:
            self._config.connection.reset_signal.register(Slot(self))
            self._config.connection.modify_epp_signal.register(Slot(lambda x: self(EppReset())))

        self._logger = logging.getLogger(type(self).__name__)

        self._thread: EvalThread | None = None  # type: ignore
        self._running = True
        self._condition = Condition()
        self._next_eval_config: EvalConfig | None = None  # type: ignore

    @staticmethod
    def get_validator() -> type[EvalRunnerValidator]:
        return EvalRunnerValidator

    def stop(self):
        self._running = False
        if self._thread:
            self._thread.stop()

    def _on_eval_config(self, message: EvalConfig):
        self._logger.debug("Setting next_schema")
        with self._condition:
            self._next_eval_config = message
            if self._thread:
                self._logger.debug("Stopping thread")
                self._thread.stop()

    def _on_epp_reset(self):
        self._logger.debug("Restarting thread...")
        with self._condition:
            if self._thread:
                self._thread.reset()

    def __call__(self, message: EvalConfig | EppReset):
        if isinstance(message, EvalConfig):
            return self._on_eval_config(message)

        if isinstance(message, EppReset):
            return self._on_epp_reset()

        raise RuntimeError(f"Unhandled message type {type(message)}")

    def _run_loop(self, output_dir: Path | None = None):
        count = 0
        while self._running:
            if count >= 1 and not self._config.run_forever:
                break

            if self._next_eval_config:
                with self._condition:
                    self._logger.debug("Get lock")
                    if self._next_eval_config:
                        count += 1

                        if output_dir:
                            self._next_eval_config.experiment.engine.rllib.output_dir = output_dir

                        self._logger.debug("Creating new EvalThread")
                        self._thread = EvalThread(config={"eval_config": self._next_eval_config}, name=f"{type(self).__name__}{count}")
                        self._next_eval_config = None
                        self._thread.start()
                if self._thread:
                    self._logger.debug("Thread join...")
                    self._thread.join()
                    del self._thread
                    gc.collect()
                    self._logger.debug("Thread joined")

            else:
                self._logger.debug("Waiting for next_schema...")
                time.sleep(1)

    def run(self):
        if self._config.tmpdir_base:
            os.makedirs(self._config.tmpdir_base, exist_ok=True)
            with tempfile.TemporaryDirectory(dir=self._config.tmpdir_base) as output_dir:
                self._run_loop(Path(output_dir))
        else:
            self._run_loop()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

import abc

from corl.libraries.plugin_library import PluginLibrary
from corl.libraries.property import BoxProp, DiscreteProp, MultiBinary
from corl.libraries.units import Quantity
from corl.parts.comms import Comms, CommsValidator
from corl.simulators.base_parts import BasePlatformPartValidator


class MemoryStoreValidator(CommsValidator):
    """Validator for memory store"""


class MemoryStore(Comms, abc.ABC):
    """Platform part for sharing intra-platform data (i.e. data between agents on the same platform)"""

    def __init__(self, parent_platform, config, property_class) -> None:
        self.config: MemoryStoreValidator
        super().__init__(parent_platform, config, property_class)

        self._data = self.default_value

    @staticmethod
    def get_validator() -> type[BasePlatformPartValidator]:
        """
        return the validator that will be used on the configuration
        of this part
        """
        return MemoryStoreValidator

    def _set_data(self, data: Quantity) -> None:
        self._data = data

    def _get_data(self) -> Quantity:
        return self._data


PluginLibrary.AddClassToGroup(MemoryStore.embed_properties(BoxProp), "BoxPropMemoryStore", {})
PluginLibrary.AddClassToGroup(MemoryStore.embed_properties(MultiBinary), "MultiBinaryPropStore", {})
PluginLibrary.AddClassToGroup(MemoryStore.embed_properties(DiscreteProp), "DiscretePropStore", {})



"""
---------------------------------------------------------------------------


Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Transforms the raw observations from the ACT3MultiAgentEnv into the transformed
format a policy expects.
"""

import copy
import dataclasses
import typing
from collections import OrderedDict
from pathlib import Path

import gymnasium
import numpy as np
import torch
from flatten_dict import flatten_dict
from pydantic import BaseModel, ConfigDict
from ray.rllib.algorithms.ppo.ppo_torch_policy import PPOTorchPolicy
from ray.rllib.evaluation.collectors.agent_collector import AgentCollector
from ray.rllib.models.modelv2 import restore_original_dimensions
from ray.rllib.models.preprocessors import get_preprocessor
from ray.rllib.models.repeated_values import RepeatedValues
from ray.rllib.policy import Policy
from ray.rllib.policy.sample_batch import SampleBatch, concat_samples
from ray.rllib.utils.filter import Filter
from ray.rllib.utils.spaces.repeated import Repeated
from ray.rllib.utils.typing import AgentID

from corl.environment.multi_agent_env import ACT3MultiAgentEnv
from corl.evaluation.episode_artifact import EpisodeArtifact, Preprocessor


@dataclasses.dataclass
class AgentCheckpoint:
    """Agent Checkpoint and Name"""

    agent_name: AgentID
    checkpoint_dir: str | Path


class ACT3MultiAgentEnvToPolicyNetworkValidator(BaseModel):
    """Validator for Network Explainability"""

    agent_checkpoints: list[AgentCheckpoint]
    env_config: dict[str, typing.Any]
    model_config = ConfigDict(arbitrary_types_allowed=True)


class ACT3MultiAgentEnvToPolicyNetworkInputs:
    """Given an environment configuration and agent checkpoints,
    takes the necessary steps to transform the raw environment observations
    into the policy network inputs.

    Parameters
    ----------
    agent_checkpoints : typing.List[AgentCheckpoint]
        A list of AgentCheckpoints
    env_config : typing.Union[EnvContext, typing.Dict]
        The environment configuration used to recreate the ACT3MultiAgentEnv
    """

    def __init__(self, **kwargs) -> None:
        self.config: ACT3MultiAgentEnvToPolicyNetworkValidator = self.get_validator()(**kwargs)
        self.env = ACT3MultiAgentEnv(self.config.env_config)
        self.policies: dict[AgentID, PPOTorchPolicy] = load_checkpoints(self.config.agent_checkpoints)
        self._policy_input_observation_names: list[str] = []
        self.discrete_inputs_indicator: np.ndarray = np.array([])

    @staticmethod
    def get_validator() -> type[ACT3MultiAgentEnvToPolicyNetworkValidator]:
        """Gets validator for class"""
        return ACT3MultiAgentEnvToPolicyNetworkValidator

    @property
    def policy_input_observation_names(self):
        """A list of strings indicating each network input"""
        return self._policy_input_observation_names

    @policy_input_observation_names.setter
    def policy_input_observation_names(self, value):
        """Custom setter that checks to ensure the expected policy inputs have not
        changed between episode artifacts"""
        if self._policy_input_observation_names:
            if self._policy_input_observation_names != value:
                raise ValueError(
                    "Network outputs changed between episoide artifacts.", f" {self._policy_input_observation_names} -> {value}"
                )
        else:
            self._policy_input_observation_names = value

    def _get_preprocessor_and_filter(
        self,  # noqa: PLR6301
        episode_artifact: EpisodeArtifact,
    ) -> tuple[dict[AgentID, Preprocessor], dict[AgentID, Filter]]:
        """Returns the policy preprocessor and filter. When `sanity_check` is True, will loop through
        preprocessors and filters to ensure they are the same throughout the episode artifact"""
        policy_preprocessor: dict[AgentID, Preprocessor] = {}
        policiy_filter: dict[AgentID, Filter] = {}
        for policy_id, policy_artifact in episode_artifact.policy_artifact.items():
            policy_preprocessor[policy_id] = policy_artifact.preprocessor
            policiy_filter[policy_id] = policy_artifact.filters
        return policy_preprocessor, policiy_filter

    def get_raw_observations(
        self, agent_id: AgentID, episode_artifact: EpisodeArtifact, include_initial_state: bool = True  # noqa: PLR6301
    ) -> list[OrderedDict[typing.Any, typing.Any]]:
        """Returns the raw observations for a policy from the episode artifact

        Parameters
        ----------
        agent_id : AgentID
            The agent id to grab the observations for.
        episode_artifact : EpisodeArtifact
            The episode artifact to get the observations from.
        include_initial_state : bool
            When True, includes the initial state of the episode. By default, True.

        Returns
        -------
        typing.List[typing.Dict[str, typing.Any]]
            A list of raw observations.
        """
        raw_observations: list[OrderedDict] = []

        # The initial state is not available from the simulator steps
        # and, therefore, we (optionally) append it here.
        if include_initial_state:
            agent_initial_state = episode_artifact.initial_state[agent_id]
            raw_observations.append(agent_initial_state)
        # Grabs the names of the observations that the policy network expects

        for step in episode_artifact.steps:
            agent_step: EpisodeArtifact.AgentStep = step.agents[agent_id]
            agent_obs: OrderedDict[typing.Any, typing.Any] = agent_step.observations
            # Removes any observations that were present in the environment and excluded from training
            raw_observations.append(agent_obs)
        return raw_observations

    def _get_policy_input_observation_names(self, input_obs: np.ndarray, preprocessor: Preprocessor) -> list[str]:
        """Returns the policy input observation names and sets `_policy_input_observation_names` and
        `discrete_inputs_indicator`.

        Parameters
        ----------
        agent_id : AgentID
            The Agent ID corresponding to the policy to use
        input_obs : np.ndarray
            An array representing an example policy input sample.

        Returns
        -------
        typing.List[str]
            A list of policy input observation names
        """
        assert len(input_obs.shape) == 1, "The `input_obs` is assumed to be a flattened array of shape num_input_features."
        # NOTE: The order of the observation space as its listed in the OrderedDict matters. We use the one that is an attribute
        # of the preprocessor itself, since the preprocessor handles the transformation. We have had issues with the observation space
        # found in the loaded checkpoint are ordered differently from the observation space found in the preprocessor.

        # This will output a dict of arrays (pre-flattened arrays)
        _original_space_dict = preprocessor.observation_space.original_space
        ref_idx_dict = restore_original_dimensions(
            np.expand_dims(np.arange(0, len(input_obs)), axis=0), obs_space=preprocessor.observation_space, tensorlib=np
        )
        spaces_dict_restored_dims = restore_original_dimensions(
            np.expand_dims(input_obs, axis=0), obs_space=preprocessor.observation_space, tensorlib=np
        )
        # Flattens various dictionaries that keep track of spaces and indices
        flattened_spaces_dict = _to_flattened_dict(input_dict=spaces_dict_restored_dims)
        flattened_original_spaces_dict = _to_flattened_dict(input_dict=_original_space_dict)
        ref_idx_dict = _to_flattened_dict(input_dict=ref_idx_dict)

        policy_input_names = [""] * len(input_obs)
        discrete_value_indicator = 1
        discrete_values = np.zeros_like(input_obs)
        for key in flattened_spaces_dict:
            obs_tuple = tuple(name for name in _flatten_tuples(key) if name != "direct_observation")
            obs_name = ".".join(obs_tuple)
            space_ = flattened_original_spaces_dict[key]
            _idx_array = ref_idx_dict[key].flatten()
            # When the space is either Discrete or MultiDiscrete, add a non-zero value to the
            # discrete inputs indicator. All one-hot encoded vectors representing the same feature are represented
            # by the same int value. This is used later to consolidate the one-hot encoded vectors for visualization
            if isinstance(space_, gymnasium.spaces.MultiDiscrete | gymnasium.spaces.Discrete):
                discrete_values[_idx_array] = discrete_value_indicator
                discrete_value_indicator += 1
            for grouped_idx, _idx in enumerate(_idx_array):
                policy_input_names[_idx] = f"{obs_name}_{grouped_idx}"
        self._policy_input_observation_names = policy_input_names
        self.discrete_inputs_indicator = discrete_values
        return policy_input_names

    def _get_transformed_obs(
        self,  # noqa: PLR6301
        prepped_obs: dict[str, typing.Any],
        preprocessor: Preprocessor,
        policy_filter: typing.Any,
    ) -> np.ndarray:
        """Given a raw observation and a policy-id, transforms the raw observations into a numpy array.

        Parameters
        ----------
        prepped_obs : typing.Dict[str, typing.Any]
            The observation prepared for training
        preprocessor : Preprocessor
            The policy preprocessor
        filter : Filter
            The policy filter(s)

        Returns
        -------
        np.ndarray
            The transformed observations.
        """
        if preprocessor is not None:
            transformed_obs = preprocessor.transform(prepped_obs)
        return policy_filter(transformed_obs)

    @staticmethod
    def _del_keys(raw_obs: dict[str, typing.Any], remove_keys: set):
        for key in remove_keys:
            del raw_obs[key]
        return raw_obs

    def raw_observations_to_sample_batch(
        self,
        agent_id: AgentID,
        episode_artifact: EpisodeArtifact,
        include_initial_state: bool = True,
    ) -> SampleBatch:
        """Converts the raw observations to the expected policy network input format.
        This includes any normalizations or transformations that are configured to happen by
        the environment.

        Parameters
        ----------
        agent_id : AgentID
            A string indicating the agent-id
        episode_artifact: EpisodeArtifact
            Episode artifact to process
        include_initial_state: bool
            When True, includes the initial state (timestep 0).

        Returns
        -------
        SampleBatch
            Sample batch of observations.
        """
        batch_obs = []
        raw_observations = copy.deepcopy(
            self.get_raw_observations(agent_id=agent_id, episode_artifact=episode_artifact, include_initial_state=include_initial_state)
        )
        policy_artifact: EpisodeArtifact.PolicyArtifact = episode_artifact.policy_artifact[agent_id]
        _agent = self.env.agent_dict[agent_id]
        for idx, obs in enumerate(raw_observations):
            array_obs = self._get_transformed_obs(
                _agent.create_training_observations(obs),
                preprocessor=get_preprocessor(self.env.observation_space[agent_id])(self.env.observation_space[agent_id]),  # type: ignore
                policy_filter=policy_artifact.filters,
            )
            batch_obs.append(array_obs)
            if idx == 0:
                self._get_policy_input_observation_names(
                    input_obs=array_obs,
                    preprocessor=get_preprocessor(self.env.observation_space[agent_id])(  # type: ignore
                        self.env.observation_space[agent_id]  # type: ignore
                    ),
                )
        view_requirements = self.policies[agent_id].model.view_requirements
        sample_batch_list = []

        collector = AgentCollector(view_reqs=view_requirements, max_seq_len=1, is_training=False)
        for idx, one_batch in enumerate(batch_obs):
            # Adds an observation to the buffer, the rest of the inputs are dummy variables
            collector.add_init_obs(episode_id=0, agent_index=0, env_id=0, init_obs=one_batch, t=idx)
            sample_batch_list.append(collector.build_for_inference())
        final_sample_batch = concat_samples(sample_batch_list)
        return SampleBatch({key: torch.from_numpy(array_value) for key, array_value in final_sample_batch.items()})


def _flatten_tuples(input_tuple: tuple):
    for value in input_tuple:
        if isinstance(value, tuple):
            yield from _flatten_tuples(value)
        else:
            yield value


def _to_flattened_dict(input_dict: dict[typing.Any, typing.Any], reducer: str = "tuple"):
    """Helper function to turn all nested dicts and repeated values into flattened dictionaries -
    when no repeated values are found, the input dict will still be flattened"""
    while any(isinstance(_value, dict | gymnasium.spaces.Dict | RepeatedValues | Repeated) for _, _value in input_dict.items()):
        new_dict = {}
        for key, value in input_dict.items():
            if isinstance(key, tuple):
                key = tuple(_flatten_tuples(key))  # noqa: PLW2901
            if isinstance(value, RepeatedValues):
                new_dict[key] = value.values
            elif isinstance(value, Repeated):
                new_dict[key] = value.child_space
            else:
                new_dict[key] = value
        input_dict = flatten_dict.flatten(new_dict, reducer=reducer)
    return flatten_dict.flatten(input_dict, reducer=reducer)


def load_checkpoints(checkpoints: list[AgentCheckpoint]) -> dict[AgentID, PPOTorchPolicy]:
    """Loads a checkpoint

    Parameters
    ----------
    checkpoints : typing.List[AgentCheckpoint]
        A list of agent checkpoints.

    Returns
    -------
    typing.Dict[AgentID, Policy]
        Loaded checkpoints
    """
    # By default, the checkpoint is loaded onto the device it was trained on.
    # To avoid mismatches between the model and the tensors later on, we load
    # the model onto the CPU.
    DEVICE = "cpu"
    loaded_checkpoints = {}
    for agent_checkpoint in checkpoints:
        _policy = Policy.from_checkpoint(agent_checkpoint.checkpoint_dir)
        # Set the device
        if isinstance(_policy, Policy):
            _policy.devices = [DEVICE]
            _policy.device = DEVICE

            _policy.model = _policy.model.to(DEVICE)
            loaded_checkpoints[agent_checkpoint.agent_name] = _policy
        else:
            _policy[agent_checkpoint.agent_name].devices = [DEVICE]
            _policy[agent_checkpoint.agent_name].device = DEVICE

            _policy[agent_checkpoint.agent_name].model = _policy[agent_checkpoint.agent_name].model.to(DEVICE)
            loaded_checkpoints[agent_checkpoint.agent_name] = _policy[agent_checkpoint.agent_name]
    return loaded_checkpoints
